#include "StdAfx.h"
#include "emule.h"
#include "barshader.h"
#include "math.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#ifndef M_PI_2
#define M_PI_2     1.57079632679489661923
#endif

// Barry
#ifndef PI
#define PI 3.14159265358979323846264338328
#endif

#define HALF(X) (((X) + 1) / 2)

CBarShader::CBarShader(uint32 height, uint32 width) 
{
	m_iWidth = width;
	m_iHeight = height;
	m_dwFileSize = 1;
	m_Spans.SetAt(0, 0);
	m_pdblModifiers = NULL;
	m_bIsPreview=false;
	m_used3dlevel = 0;
}

CBarShader::~CBarShader(void) {
	delete[] m_pdblModifiers;
	m_Spans.RemoveAll();
}

void CBarShader::Reset() {
	Fill(0);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CBarShader::BuildModifiers()
{
	if (m_pdblModifiers != NULL)
	{
		delete[] m_pdblModifiers;
		m_pdblModifiers = NULL; // 'new' may throw an exception
	}

	if (!m_bIsPreview) 
		m_used3dlevel = g_eMuleApp.m_pGlobPrefs->Get3DDepth();

	// Barry - New property page slider to control depth of gradient

	// Depth must be at least 2
	// 2 gives greatest depth, the higher the value, the flatter the appearance
	// m_pdblModifiers[count-1] will always be 1, m_pdblModifiers[0] depends on the value of depth
	double	depths[5] = { 1.5, 3.0, 4.0, 4.50, 4.75 };		//Cax2 aqua bar - smoother gradient jumps...
	double	depth = 7 - depths[((m_used3dlevel>10)?256-m_used3dlevel:m_used3dlevel) -1];		//Cax2 aqua bar
	int count = HALF(m_iHeight);
	double piOverDepth = PI/depth;
	double base = piOverDepth * (depth / 2 - 1);
	double increment = piOverDepth / (count - 1);

	m_pdblModifiers = new double[count];
	for (int i = 0; i < count; i++)
		m_pdblModifiers[i] = static_cast<double>(sin(base + i * increment));
}

void CBarShader::SetWidth(int width) {
	if(m_iWidth != width) {
		m_iWidth = width;
		if (m_dwFileSize)
			m_dblPixelsPerByte = (double)m_iWidth / m_dwFileSize;
		else
			m_dblPixelsPerByte = 0.0;
		if (m_iWidth)
			m_dblBytesPerPixel = (double)m_dwFileSize / m_iWidth;
		else
			m_dblBytesPerPixel = 0.0;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CBarShader::SetFileSize(uint32 dwFileSize)
{
	if (m_dwFileSize != dwFileSize)
	{
		m_dwFileSize = dwFileSize;
		if (m_dwFileSize)
			m_dblPixelsPerByte = static_cast<double>(m_iWidth) / m_dwFileSize;
		else
			m_dblPixelsPerByte = 0.0;
		if (m_iWidth)
			m_dblBytesPerPixel = static_cast<double>(m_dwFileSize) / m_iWidth;
		else
			m_dblBytesPerPixel = 0.0;
	}
}

void CBarShader::SetHeight(int height)
{
	if(m_iHeight != height)
 	{
		m_iHeight = height;

		BuildModifiers();
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CBarShader::FillRange(uint32 start, uint32 end, COLORREF crColor)
{
	if(end > m_dwFileSize)
		end = m_dwFileSize;

	if(start >= end)
		return;
	POSITION endpos = m_Spans.FindFirstKeyAfter(end+1);

	if (endpos)
		m_Spans.GetPrev(endpos);
	else
		endpos = m_Spans.GetTailPosition();

	ASSERT(endpos != NULL);

	COLORREF endcolor = m_Spans.GetValueAt(endpos);
	endpos = m_Spans.SetAt(end, endcolor);

	for (POSITION pos = m_Spans.FindFirstKeyAfter(start+1); pos != endpos; ) {
		POSITION pos1 = pos;
		m_Spans.GetNext(pos);
		m_Spans.RemoveAt(pos1);
	}
	
	m_Spans.GetPrev(endpos);

	if (m_Spans.GetValueAt(endpos) != crColor)
		m_Spans.SetAt(start, crColor);
}

void CBarShader::Fill(COLORREF crColor) {
	m_Spans.RemoveAll();
	m_Spans.SetAt(0, crColor);
	m_Spans.SetAt(m_dwFileSize, 0);
}

void CBarShader::DrawPreview(CDC* dc, int iLeft, int iTop, uint8 previewLevel)		//Cax2 aqua bar
{
	m_bIsPreview=true;
	m_used3dlevel = previewLevel;
	BuildModifiers();
	Draw( dc, iLeft, iTop, (previewLevel==0));
	m_bIsPreview=false;
}

void CBarShader::Draw(CDC* dc, int iLeft, int iTop, bool bFlat) {
	POSITION pos = m_Spans.GetHeadPosition();
	RECT rectSpan;
	rectSpan.top = iTop;
	rectSpan.bottom = iTop + m_iHeight;
	rectSpan.right = iLeft;

	int iBytesInOnePixel = static_cast<int>(m_dblBytesPerPixel + 0.5);
	uint32 start = 0;//bsCurrent->m_dwStartOffset;
	COLORREF m_crColor = m_Spans.GetValueAt(pos);
	m_Spans.GetNext(pos);
	while(pos != NULL && rectSpan.right < (iLeft + m_iWidth)) {
		uint32 uSpan = m_Spans.GetKeyAt(pos) - start;
		int iPixels = static_cast<int>(uSpan * m_dblPixelsPerByte + 0.5);
		if(iPixels > 0) {
			rectSpan.left = rectSpan.right;
			rectSpan.right += iPixels;
			FillRect(dc, &rectSpan, m_crColor, bFlat);

			start += static_cast<int>(iPixels * m_dblBytesPerPixel + 0.5);
		} else {
			double dblRed = 0;
			double dblGreen = 0;
			double dblBlue = 0;
			uint32 iEnd = start + iBytesInOnePixel;
			int iLast = start;
			do {
				double	dblWeight = (min(m_Spans.GetKeyAt(pos), iEnd) - iLast) * m_dblPixelsPerByte;
				dblRed   += GetRValue(m_crColor) * dblWeight;
				dblGreen += GetGValue(m_crColor) * dblWeight;
				dblBlue  += GetBValue(m_crColor) * dblWeight;
				if(m_Spans.GetKeyAt(pos) > iEnd)
					break;
				iLast = m_Spans.GetKeyAt(pos);
				m_crColor = m_Spans.GetValueAt(pos);
				m_Spans.GetNext(pos);
			} while(pos != NULL);
			rectSpan.left = rectSpan.right;
			rectSpan.right++;
			FillRect(dc, &rectSpan, dblRed, dblGreen, dblBlue, bFlat);
			start += iBytesInOnePixel;
		}
		while(pos != NULL && m_Spans.GetKeyAt(pos) < start) {
			m_crColor = m_Spans.GetValueAt(pos);
			m_Spans.GetNext(pos);
		}
	}
}

void CBarShader::FillRect(CDC *dc, LPRECT rectSpan, COLORREF crColor, bool bFlat) {
	if(!crColor || bFlat)
	{
		CBrush gdiBrush(crColor);
		CBrush* pOldBrush = dc->SelectObject(&gdiBrush);	//eklmn: select a new brush
		dc->FillRect(rectSpan, &gdiBrush);
		dc->SelectObject(pOldBrush);		//eklmn: recover an old brush
	}
	else
		FillRect(dc, rectSpan, GetRValue(crColor), GetGValue(crColor), GetBValue(crColor), false);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CBarShader::FillRect(CDC *dc, LPRECT rectSpan, double dblRed, double dblGreen, double dblBlue, bool bFlat) 
{
	CBrush		   *pOldBrush;
	if (bFlat) 
	{
		COLORREF crColor = RGB(static_cast<int>(dblRed + .5), static_cast<int>(dblGreen + .5), static_cast<int>(dblBlue + .5));

		CBrush		gdiBrush(crColor);
		pOldBrush = dc->SelectObject(&gdiBrush);
		dc->FillRect(rectSpan, &gdiBrush);
		dc->SelectObject(pOldBrush);
	} 
	else 
	{
		if ((m_used3dlevel != g_eMuleApp.m_pGlobPrefs->Get3DDepth() && !m_bIsPreview) || m_pdblModifiers == NULL)
			BuildModifiers();

		RECT		rect;

		memcpy2(&rect, rectSpan, sizeof(RECT));

		int		iTop = rect.top;
		int		iBot = rect.bottom;
		int		iMax = HALF(m_iHeight);
		double	dblRed1,dblRed2,dblBlue1,dblBlue2,dblGreen1,dblGreen2;

		if (m_used3dlevel > 10)		//Cax2 aqua bar
		{
			dblBlue2 = 1.0 - .025 * (256 - m_used3dlevel);		//variable central darkness - from 97.5% to 87.5% of the original colour...

			dblRed1 = dblGreen1 = dblBlue1 = 255;

			//dblRed1=(255 + dblRed)/2;

			dblRed2 = dblBlue2 * dblRed - dblRed1;

			//dblGreen1=(255 + dblGreen)/2;

			dblGreen2 = dblBlue2 * dblGreen - dblGreen1;

			//dblBlue1=(255 + dblBlue)/2;

			dblBlue2 = dblBlue2 * dblBlue - dblBlue1;
		}
		else
		{
			dblRed1 = dblGreen1 = dblBlue1 = .5;
			dblRed2 = dblRed;
			dblGreen2 = dblGreen;
			dblBlue2 = dblBlue;
		}
		for (int i = 0; i < iMax; i++) 
		{
			CBrush		cbNew(RGB( static_cast<int>(dblRed1 + dblRed2*m_pdblModifiers[i]),
								   static_cast<int>(dblGreen1 + dblGreen2*m_pdblModifiers[i]),
								   static_cast<int>(dblBlue1 + dblBlue2*m_pdblModifiers[i]) ));

			pOldBrush = dc->SelectObject(&cbNew);
			rect.top = iTop + i;
			rect.bottom = iTop + i + 1;
			dc->FillRect(&rect, &cbNew);

			rect.top = iBot - i - 1;
			rect.bottom = iBot - i;
			dc->FillRect(&rect, &cbNew);
			dc->SelectObject(pOldBrush);
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
