//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "ClientList.h"
#include "emule.h"
#include "otherfunctions.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CClientList::CClientList()
{
	m_pvecDirtyClients = new ClientVector();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CClientList::~CClientList()
{
	delete m_pvecDirtyClients;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	xrmb : statsclientstatus
void CClientList::GetStatistics(uint32 &totalclient, uint32 stats[], uint32 clientsPorts[], ClientsData *pData, uint32 &totalMODs, CMap<POSITION, POSITION, uint32, uint32> *MODs)
{
	EMULE_TRY
	if (!pData )
		return ;

	POSITION pos1, pos2, pos_MOD;
	CUpDownClient* cur_client;
//	CMap<POSITION,POSITION, uint32, uint32> MODs;
	CString MOD_name;
	uint32 MOD_count;

	totalclient = m_clientList.GetCount();

	for (int i = 0;i < SO_LAST;i++)
	{
		pData->m_pClients[i].RemoveAll();
		stats[i] = 0;
	}
	pData->m_pClients[i].RemoveAll(); // reset the last one

//	reset ports information
	clientsPorts[0] = 0;
	clientsPorts[1] = 0;

//reset MODs variables
	totalMODs = 0;
	MODs->RemoveAll();

	for (pos1 = m_clientList.GetHeadPosition();( pos2 = pos1 ) != NULL;)
	{
		m_clientList.GetNext(pos1);
		cur_client = m_clientList.GetAt(pos2);
		int version	= 0;

		switch (cur_client->GetClientSoft())
		{
			case SO_UNKNOWN :
				stats[SO_UNKNOWN]++;
				break;
			case SO_EDONKEY :
				stats[SO_EDONKEY]++;
				(pData->m_pClients[SO_EDONKEY])[cur_client->GetVersion()]++;
				break;
			case SO_EDONKEYHYBRID :
				stats[SO_EDONKEYHYBRID]++;
				(pData->m_pClients[SO_EDONKEYHYBRID])[cur_client->GetVersion()]++;
				break;
			case SO_EMULE :
				stats[SO_EMULE]++;
				version = cur_client->GetMuleVersion();
				(pData->m_pClients[SO_EMULE])[version]++;
				break;
			case SO_OLDEMULE:
				stats[SO_OLDEMULE]++;
				version = cur_client->GetMuleVersion();
				(pData->m_pClients[SO_OLDEMULE])[version]++;
				break;
			case SO_CDONKEY :  //Didn't get much time to test this "eMule Compatable" feature, so for this version they will just be added to our total.
				stats[SO_CDONKEY]++;
				version = cur_client->GetMuleVersion();
				(pData->m_pClients[SO_CDONKEY])[version]++;
				break;
			case SO_SHAREAZA:
				stats[SO_SHAREAZA]++;
				version = cur_client->GetMuleVersion();
				(pData->m_pClients[SO_SHAREAZA])[version]++;
				break;
			case SO_LMULE:
				stats[SO_LMULE]++;
				version = cur_client->GetMuleVersion();
				(pData->m_pClients[SO_LMULE])[version]++;
				break;

			case SO_MLDONKEY:
				stats[SO_MLDONKEY]++;
				(pData->m_pClients[SO_MLDONKEY])[version]++;
				break;
		}

		(pData->m_pClients[SO_LAST])[cur_client->GetDownloadState()]++;

		switch (cur_client->GetUserPort())
		{
			case 4662:
				++clientsPorts[0]; // Default Port
				break;
			default:
				++clientsPorts[1]; // Other Port
		}

	//	eklmn: count a MODs
		MOD_name = cur_client->GetModVersion();
		if (!MOD_name.IsEmpty())
		{
			totalMODs++;
			pos_MOD = liMODsTypes.Find(CString(MOD_name));
			if (!pos_MOD)
			{
				pos_MOD = liMODsTypes.AddTail(MOD_name);
				MODs->SetAt(pos_MOD, 1 );
			}
			else
			{
				MOD_count = 0;
				MODs->Lookup(pos_MOD, MOD_count);
				MOD_count++;
				MODs->SetAt(pos_MOD, MOD_count);
			}
			MOD_name.Empty();
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CClientList::AddClient(CUpDownClient* toadd, bool bSkipDupTest)
{
	EMULE_TRY

	if (toadd == NULL)
		return false;

	if ( !bSkipDupTest)
	{
		if (m_clientList.Find(toadd))
			return false;
	}
	if (m_pctlClientList != NULL && m_pctlClientList->m_hWnd != NULL)
		m_pctlClientList->AddClient(toadd);
	m_clientList.AddTail(toadd);

	return true;

	EMULE_CATCH
		
	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientList::RemoveClient(CUpDownClient *pClient)
{
	EMULE_TRY

	POSITION		pos = m_clientList.Find(pClient);

	if (pos != NULL)
	{
	//	g_eMuleApp.AddDebugLogLine(false,_T("%s: Removing client %s addr(0x%08X) from client list"),__FUNCTION__,pClient->GetUserName(),pClient);

	//	Just to be sure...
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(pClient, ETS_DISCONNECT);
		g_eMuleApp.m_pUploadQueue->RemoveFromWaitingQueue(pClient);
		g_eMuleApp.m_pDownloadQueue->RemoveSource(pClient);
		m_pctlClientList->RemoveClient(pClient);

	//	g_eMuleApp.AddDebugLogLine(false,_T("%s: REALLY removing client %s addr(0x%08X) from client list"),__FUNCTION__,pClient->GetUserName(),pClient);

		m_clientList.RemoveAt(pos);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientList::DeleteAll()
{
	EMULE_TRY

	g_eMuleApp.m_pUploadQueue->DeleteAll();
	g_eMuleApp.m_pDownloadQueue->DeleteAll();
	POSITION pos1, pos2;
	for (pos1 = m_clientList.GetHeadPosition();( pos2 = pos1 ) != NULL;)
	{
		m_clientList.GetNext(pos1);
		CUpDownClient* cur_client = m_clientList.GetAt(pos2);
		m_clientList.RemoveAt(pos2);
		if (cur_client)
			delete cur_client; // recursiv: this will call RemoveClient
	}
	liMODsTypes.RemoveAll();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef OLD_SOCKETS_ENABLED
bool CClientList::AttachToAlreadyKnown(CUpDownClient **ppClient, CClientReqSocket *pSocket)
{
	EMULE_TRY

	POSITION			pos1, pos2;
	CUpDownClient	   *pClient = (*ppClient);

	for (pos1 = m_clientList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_clientList.GetNext(pos1);

		CUpDownClient		*pClient2 = m_clientList.GetAt(pos2);

	//	I dont know how and why this can happen... but emule crashed here all the time for me!
		if (pClient2 != NULL && pClient != pClient2 && pClient->Compare(pClient2) != 0)
		{
			ASSERT ( (pClient != pClient2) );
			if (pSocket != NULL)
			{
				if (pClient2->m_pRequestSocket != NULL)
				{
					pClient2->m_pRequestSocket->Safe_Delete();
				}
				pClient2->m_pRequestSocket = pSocket;
				pClient->m_pRequestSocket = NULL;
			}
			*ppClient = NULL;
			delete pClient;
			*ppClient = pClient2;
			return true;
		}
	}

	EMULE_CATCH

	return false;
}
#endif //OLD_SOCKETS_ENABLED
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CClientList::FindClientByIP(uint32 clientip, uint16 port)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = m_clientList.GetHeadPosition();( pos2 = pos1 ) != NULL;)
	{
		m_clientList.GetNext(pos1);
		CUpDownClient* cur_client = m_clientList.GetAt(pos2);
		if (cur_client)
			if (cur_client->GetIP() == clientip && cur_client->GetUserPort() == port)
				return cur_client;
	}

	EMULE_CATCH

	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CClientList::FindClientByUserHash(uchar* clienthash)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = m_clientList.GetHeadPosition();( pos2 = pos1 ) != NULL;)
	{
		m_clientList.GetNext(pos1);
		CUpDownClient* cur_client = m_clientList.GetAt(pos2);
		if (cur_client)
			if (!md4cmp(cur_client->GetUserHash() , clienthash))
				return cur_client;
	}

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CUpDownClient* CClientList::FindClientByPortAndID(uint32 dwUserID,uint16 nPort)
{
	EMULE_TRY

	POSITION pos1, pos2;
	CUpDownClient* pClient;
	for (pos1 = m_clientList.GetHeadPosition();( pos2 = pos1 ) != NULL;)
	{
		m_clientList.GetNext(pos1);
		pClient = m_clientList.GetAt(pos2);
		if(pClient)
			if (pClient->GetUserID() == dwUserID && pClient->GetUserPort() == nPort)
				return pClient;
	}
	
	EMULE_CATCH

	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifdef OLD_SOCKETS_ENABLED
void CClientList::Debug_SocketDeleted(CClientReqSocket *pSocket)
{
	EMULE_TRY

	POSITION		pos1, pos2;

	for (pos1 = m_clientList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_clientList.GetNext(pos1);

		CUpDownClient		*pClient = m_clientList.GetAt(pos2);

		if (!AfxIsValidAddress(pClient, sizeof(CUpDownClient)))
		{
			AfxDebugBreak();
		}
		if (pClient && pClient->m_pRequestSocket == pSocket)
		{
		//	AfxDebugBreak();
		}
	}

	EMULE_CATCH
}
#endif //OLD_SOCKETS_ENABLED 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CClientList::IsValidClient(CUpDownClient *pClient)
{
	EMULE_TRY

	if (pClient == NULL)
		return false;

	return m_clientList.Find(pClient);

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	[i_a]
void CClientList::ResetAllValueCaches()
{
	EMULE_TRY

	if (!m_clientList.IsEmpty()) // Fix reCDVst
		for (POSITION pos = m_clientList.GetHeadPosition(); pos != 0; m_clientList.GetNext(pos))
		{
			CUpDownClient * cur_client = m_clientList.GetAt(pos);
			ASSERT(cur_client);
			if (cur_client)
				cur_client->ResetValueCache();
		}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//netwolf: complete sources (zegzav) 13.05.03 - BEGIN
void CClientList::GetClientListByFileID(CTypedPtrList<CPtrList, CUpDownClient*> *m_pClientList, const uchar *fileid)
{
	EMULE_TRY

	m_pClientList->RemoveAll();
	for (POSITION pos = m_clientList.GetHeadPosition(); pos != 0; )
	{
		CUpDownClient *cur_src = m_clientList.GetNext(pos);
		if (md4cmp(cur_src->GetUploadFileID(), fileid) == 0)
			m_pClientList->AddTail(cur_src);
	}

	EMULE_CATCH
}
//netwolf: complete sources (zegzav) 13.05.03 - END
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//eklmn: count A4AF sources
uint32 CClientList::GetA4AFSourcesCount()
{
	EMULE_TRY

	CUpDownClient * cur_client;
	POSITION pos1, pos2;
	uint32 countA4AF = 0;

	for (pos1 = m_clientList.GetHeadPosition();( pos2 = pos1 ) != NULL;)
	{
		m_clientList.GetNext(pos1);
		cur_client = m_clientList.GetAt(pos2);
		if (!cur_client->m_otherRequestsList.IsEmpty())
			countA4AF++;
	}

	return countA4AF;

	EMULE_CATCH
	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientList::UpdateClient(CUpDownClient *pClient)
{
	if (pClient != NULL)
	{
		POSITION		posClient = m_clientList.Find(pClient);

	//	If the client is in the list...
		if (posClient != NULL)
		{
			AddDirtyClient(pClient);
		//
		//	Send a notification to the client list control
			m_pctlClientList->PostRefreshMessage();

			g_eMuleApp.m_pDownloadList->UpdateSource(pClient);
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CClientList::AddDirtyClient(CUpDownClient *pClient)
{
//	If 'pClient' is not already in the dirty client list...
	if (::find(m_pvecDirtyClients->begin(),m_pvecDirtyClients->end(),pClient) == m_pvecDirtyClients->end())
	{
		m_pvecDirtyClients->push_back(pClient);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetDirtyClients() detaches the current dirty client list and returns it to the caller (who's responsible
//		for destroying it). This allows the client list control to update without locking down the client list.
CClientList::ClientVector *CClientList::GetDirtyClients()
{
	ClientVector		*pvecDirtyClients = NULL;

	if (!m_pvecDirtyClients->empty())
	{
		pvecDirtyClients = m_pvecDirtyClients;
		m_pvecDirtyClients = new ClientVector();
	}

	return pvecDirtyClients;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
