// EngineData.cpp
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Sockets/EmEngine.h"
#include "EngineData.h"
#include "resource.h"
#include "ServerList.h"
#include "emule.h"
#include "Sockets/Tasks.h"

#ifdef NEW_SOCKETS

//////////////////////////////////////////////////////////////////////
CEngineData::CEngineData(CEmEngine& stEngine)
: m_stEngine(stEngine)
{
	m_stConnected.nConnState = SERVER_DISCONNECTED;
	m_stConnected.hSocket = NULL;
	m_stConnected.nAddr = 0;
	m_stConnected.nPort = 0;
	m_stConnected.pServer = NULL;

	m_stConnecting.nConnState = SERVER_DISCONNECTED;
	m_stConnecting.hSocket = NULL;
	m_stConnecting.nAddr = 0;
	m_stConnecting.nPort = 0;
	m_stConnecting.pServer = NULL;

	m_nClientID = 0;

	m_pServerList = NULL;
}

//////////////////////////////////////////////////////////////////////
CEngineData::~CEngineData()
{
}

//////////////////////////////////////////////////////////////////////
bool CEngineData::IsConnecting()
{
	return (m_stConnecting.nConnState == SERVER_CONNECTING || 
			m_stConnecting.nConnState == SERVER_WAITFORLOGIN); 
}

//////////////////////////////////////////////////////////////////////
bool CEngineData::IsConnected()
{ 
	return (m_stConnected.nConnState == SERVER_CONNECTED);
}

//////////////////////////////////////////////////////////////////////
bool CEngineData::IsLowID()
{ 
	return (m_nClientID < 0x1000000); 
}

//////////////////////////////////////////////////////////////////////
// Executed in GUI thread
void CEngineData::ImportServerState(CTask_ServerState* pTask)
{
	if(!pTask)
	{
		ASSERT(FALSE);
		return;
	}

	// should keep independent serverlist in the future
	if(!m_pServerList)
		m_pServerList = g_eMuleApp.m_pServerList;

	if(m_nClientID != pTask->m_nClientID)
	{
		m_nClientID = pTask->m_nClientID;
		AddLogLine(false, IDS_NEWCLIENTID, m_nClientID);
	}

	ServerState* pState = pTask->m_bForConnected ? &m_stConnected : &m_stConnecting;
	pState->hSocket = pTask->m_stState.hSocket;
	pState->nAddr = pTask->m_stState.nAddr;
	pState->nPort = pTask->m_stState.nPort;
	if(pState->pServer == NULL && m_pServerList && pState->nAddr != 0)
		pState->pServer = m_pServerList->GetServerByIP(pState->nAddr, pState->nPort);

	if(pState->nConnState != pTask->m_stState.nConnState)
	{
		pState->nConnState = pTask->m_stState.nConnState;
		switch(pState->nConnState)
		{
		case SERVER_DISCONNECTED:
			if(m_stConnected.nConnState == SERVER_DISCONNECTED)
				g_eMuleApp.m_pdlgEmule->ShowConnectionState(false);
			break;
		case SERVER_CONNECTING:
			g_eMuleApp.m_pdlgEmule->ShowConnectionState(false);
			break;
		case SERVER_WAITFORLOGIN:
			if(m_stConnecting.pServer)
				AddLogLine(false, IDS_CONNECTEDTOREQ,
					m_stConnecting.pServer->GetListName(),
					m_stConnecting.pServer->GetFullIP(),
					m_stConnecting.pServer->GetPort());
			break;
		case SERVER_CONNECTED:
			if(m_stConnected.pServer)
			{
				AddLogLine(true, IDS_CONNECTEDTO, m_stConnected.pServer->GetListName());
				g_eMuleApp.m_pdlgEmule->ShowConnectionState(true, m_stConnected.pServer->GetListName());
//				if(IsLowID())
//					AddLogLine(false, GetResString(IDS_GOTLOWID));

				// Write online signature
				g_eMuleApp.OnlineSig();

				// Resume downloading files
				g_eMuleApp.m_pDownloadQueue->ResumeFiles();

				m_stEngine.SendSharedListToServer();
			}
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////
void CEngineData::SetConnectionState(bool bForConnected, short nConnectionState)
{
	ServerState* pState = bForConnected ? &m_stConnected : &m_stConnecting;
	pState->nConnState = nConnectionState;	

	switch(nConnectionState)
	{
	case SERVER_DISCONNECTED:
		pState->hSocket = NULL;
		pState->nAddr = 0;
		pState->nPort = 0;
		pState->pServer = NULL;
		break;
	case SERVER_CONNECTING:
		break;
	case SERVER_WAITFORLOGIN:
		break;
	case SERVER_CONNECTED:
		m_stConnected.hSocket = m_stConnecting.hSocket;
		m_stConnected.nAddr = m_stConnecting.nAddr;
		m_stConnected.nPort = m_stConnecting.nPort;
		m_stConnected.pServer = m_stConnecting.pServer;
		m_stConnecting.hSocket = NULL;
		m_stConnecting.pServer = NULL;
		m_stConnected.nConnState = SERVER_CONNECTED;
		m_stConnecting.nConnState = SERVER_DISCONNECTED;
		bForConnected = true;
		break;
	}

	ExportServerState(bForConnected);
}

//////////////////////////////////////////////////////////////////////
void CEngineData::SetClientID(long cID)
{
	m_nClientID = cID; 

	ExportServerState(true);
}

//////////////////////////////////////////////////////////////////////
void CEngineData::ConnectingTo(CEmClient_Server* pClient)
{
	m_stConnecting.hSocket = pClient->m_hSocket;
//	m_stConnecting.pServer = m_pServerList->GetServerByIP(pClient->m_nAddr, pClient->m_uPort);
	m_stConnecting.nAddr = pClient->m_nAddr;
	m_stConnecting.nPort = pClient->m_uPort;
	ExportServerState(false);
}

//////////////////////////////////////////////////////////////////////
void CEngineData::ConnectedTo(CEmClient_Server* pClient)
{
	m_stConnected.hSocket = pClient->m_hSocket;
//	m_stConnected.pServer = m_pServerList->GetServerByIP(pClient->m_nAddr, pClient->m_uPort);
	m_stConnected.nAddr = pClient->m_nAddr;
	m_stConnected.nPort = pClient->m_uPort;
	ExportServerState(true);
}

//////////////////////////////////////////////////////////////////////
bool CEngineData::IsLocalServer(ULONG nAddr, USHORT uPort)
{
	if(!IsConnected())
		return false;
	if(m_stConnected.pServer)
	{
		if (m_stConnected.pServer->GetIP() == nAddr &&
			m_stConnected.pServer->GetPort() == uPort)
			return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////////////
void CEngineData::ExportServerState(bool bForConnected)
{
	CTask_ServerState* pTask = new CTask_ServerState(
		bForConnected,
		bForConnected ? m_stConnected : m_stConnecting,
		m_nClientID
		);
	m_stEngine.m_stMain.Push(pTask);
}

//////////////////////////////////////////////////////////////////////
// in the future, those functions should get data from local storage,
// not from global preferences
BYTE* CEngineData::GetUserHash()
{
	return g_eMuleApp.m_pGlobPrefs->GetUserHash();
}

LONG CEngineData::GetPort()
{
	return g_eMuleApp.m_pGlobPrefs->GetPort();
}

CString CEngineData::GetUserNick()
{
	return g_eMuleApp.m_pGlobPrefs->GetUserNick();
}
//////////////////////////////////////////////////////////////////////

#endif //NEW_SOCKETS