// EngineData.h
//
//////////////////////////////////////////////////////////////////////

#pragma once

#include "Loggable.h"

#ifdef NEW_SOCKETS

enum EnumServerStates
{
	SERVER_DISCONNECTED,
	SERVER_CONNECTING,
	SERVER_WAITFORLOGIN,
	SERVER_CONNECTED
};

class CServer;
class CServerList;
class CPreferences;
struct CTask_ServerState;
struct CEmClient_Server;

struct ServerState
{
	SOCKET		hSocket;
	ULONG		nAddr;
	USHORT		nPort;
	CServer*	pServer;
	short		nConnState;
};

// One instance of that class inside worker engine,
// second instance in GUI, synchronization between both objects
class CEngineData : public CLoggable
{
public:
	CEngineData(CEmEngine& stEngine);
	~CEngineData();

	// Server Info
	bool IsConnecting();
	bool IsConnected();
	bool IsLowID();

	void ImportServerState(CTask_ServerState* pTask);

	// Update Server Info
	void SetConnectionState(bool bForConnected, short nConnectionState);
	void SetClientID(long cID);
	void ConnectingTo(CEmClient_Server* pClient);
	void ConnectedTo(CEmClient_Server* pClient);

	SOCKET GetConnectedServer(){ return m_stConnected.hSocket; }
	SOCKET GetConnectingServer(){ return m_stConnecting.hSocket; }

	ServerState* GetConnectedState(){ return &m_stConnected; }

	long GetClientID(){ return m_nClientID; }

	bool IsLocalServer(ULONG nAddr, USHORT uPort);

	// Global preferences
	BYTE* GetUserHash();
	LONG GetPort();
	CString GetUserNick();

private:
	void ExportServerState(bool bForConnected);

private:
	CEmEngine&		m_stEngine;

	ServerState		m_stConnected;
	ServerState		m_stConnecting;

	long			m_nClientID;

	CServerList*	m_pServerList;

	// +collection of CEmClientData objects
};

class CPartFile;

// Contain all information about connecting/connected client
class CEmClientData
{
public:
	CPartFile*	GetFile() const { return m_pPartFile; }

private:
	CPartFile	*m_pPartFile;
};

#endif //NEW_SOCKETS