// This file is part of eMule 
//
///LANCAST
//
// Written by moosetea, enjoy
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// This allows lan clients to add each other as peers, it could be
// very buggy and needs alot more work.
//
// Thanks to codeproject for help with the multicast code, 
// which works but is a bit of a shambles.

// Base includes needed by this class (and every class)
#include "StdAfx.h"
#include "emule.h"

// We also need the class defintion, and specific includes
#include "LanCast.h"
#include "packets.h"
#include "ED2KLink.h"
#include "otherfunctions.h"

// CLanCast

CLanCast::CLanCast(CPreferences* in_prefs)
{

	m_prefs = in_prefs;

	bStarted = false;
	if (GetLancastEnabled())
	{
		udp_timer = SetTimer(0,4329,10000,CLanCast::UDPTimerProc);
		Start();
	}
}

bool CLanCast::SendPacket(Packet* packet)
{

		char* pcSendBuffer = new char[packet->m_dwSize+2];

		// Standard UDP Header
		memcpy2(pcSendBuffer,packet->GetUDPHeader(),2);
		// The packet Data
		memcpy2(pcSendBuffer+2,packet->m_pcBuffer,packet->m_dwSize);

		// Send the LanCast packet
		if(m_SendSocket.SendTo(pcSendBuffer, packet->m_dwSize+10, (SOCKADDR*)&m_saHostGroup, sizeof(SOCKADDR), 0) == SOCKET_ERROR)
		{
			delete[] pcSendBuffer;
			return FALSE;
		}
		else
		{
			delete[] pcSendBuffer;
			return TRUE;
		}

}

void CLanCast::BroadcastHash(CKnownFile* cur_file){


	if (GetLancastEnabled())
		Start();
	else
		Stop();


	if (GetLancastEnabled()){

		CMemFile file(100);

		uint32 dwID = GetLancastIP();
		uint16 nPort = GetEmulePort();

		file.Write(&dwID,4);
		file.Write(&nPort,2);
		file.Write(cur_file->GetFileHash(),16);

		Packet* packet = new Packet(&file,OP_LANCASTPROT);
		packet->m_eOpcode = OP_HASH;

		SendPacket(packet);
		delete packet;	
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CLanCast::RecieveHash(char* pachPacket, uint32 nSize)
{
	CSafeMemFile offeredfiles((BYTE*)pachPacket,nSize);

	uint32 dwID = 0;
	uint16 nPort = 0; 
	uchar filehash[16];

	// Read Hash Data
	offeredfiles.Read(&dwID,4);
	offeredfiles.Read(&nPort,2);
	offeredfiles.Read(&filehash,16);

	 // Debug Info for Devs - The following lines should be removed in release
	CString buffer = FileHashToString(filehash);	
	AddDebugLogLine(false,_T("Lancast : Just Received %s"),buffer);

	CPartFile* pPartFile = g_eMuleApp.m_pDownloadQueue->GetFileByID((uchar*)filehash);

	if (pPartFile) 
	{
		// Debug Info for Devs - the following lines can be removed in release
		AddDebugLogLine(false,_T("Lancast : Adding source for %s"),buffer);

		// Create a new Updown client for this source (with no server info), userid is the LAN ip address
		//CUpDownClient* newsource = new CUpDownClient(nPort,dwID,0,0,file);

		// Add this LAN uldl client to the download queue, this also will check for duplicate sockets and "merge" them
		//g_eMuleApp.m_pDownloadQueue->CheckAndAddSource(file,newsource);
		CUpDownClient* pNewSource = g_eMuleApp.m_pDownloadQueue->CheckAndAddSource(pPartFile,dwID,nPort,0,0);
		if (pNewSource)
			pNewSource->SetUserName(GetResString(IDS_LANCAST_SOURCE));
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CLanCast::~CLanCast()
{
	Stop();
}

void CLanCast::OnReceive(int nErrorCode)
{

	char buffer[5000];
	CString addressbuffer;
	uint32 port;

	int32 length = ReceiveFrom(buffer,5000,addressbuffer,port);

	// if lancast is enabled proccess
	if (GetLancastEnabled()){

		// If this isnt a packet from ourselves
		if (inet_addr(addressbuffer) != GetLancastIP())
		{

			if ((uint8)buffer[0] == OP_LANCASTPROT && length != SOCKET_ERROR)
				ProcessPacket(buffer+2,length-2,buffer[1]);

		}

	}

}

bool CLanCast::ProcessPacket(char* packet, uint32 size, uint8 opcode){

	try{

		switch(opcode){
			case OP_HASH :{	
				
				RecieveHash(packet,size);
				break;
			}
			default:
				AddLogLine(false,IDS_LANCAST_UNK_OPCODE);
				return false;
		}

		return true;
	}
	catch(...){
		AddLogLine(false,IDS_LANCAST_ERR_MULITC);
		return false;
	}
}

void CALLBACK CLanCast::UDPTimerProc(HWND hwnd, UINT uMsg,UINT_PTR idEvent,DWORD dwTime){
		g_eMuleApp.m_pSharedFilesList->NextLANBroadcast();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLanCast Private member functions - MultiCast Network Code

BOOL CLanCast::CreateReceivingSocket(LPCTSTR strGroupIP, UINT nGroupPort)
{
	/* Create socket for receiving packets from multicast group */
	if(!Create(nGroupPort, SOCK_DGRAM, FD_READ))
		return FALSE;

	BOOL bMultipleApps = TRUE;		/* allow reuse of local port if needed */
	SetSockOpt(SO_REUSEADDR, (void*)&bMultipleApps, sizeof(BOOL), SOL_SOCKET);

	/* Fill m_saHostGroup_in for sending datagrams */
	memset2(&m_saHostGroup, 0, sizeof(m_saHostGroup));
	m_saHostGroup.sin_family = AF_INET;
	m_saHostGroup.sin_addr.s_addr = inet_addr(strGroupIP);
	m_saHostGroup.sin_port = htons((USHORT)nGroupPort);

	/* Join the multicast group */
	m_mrMReq.imr_multiaddr.s_addr = inet_addr(strGroupIP);	/* group addr */ 
	m_mrMReq.imr_interface.s_addr = htons(INADDR_ANY);		/* use default */ 
	if(setsockopt(m_hSocket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char FAR *)&m_mrMReq, sizeof(m_mrMReq)) < 0)
		return FALSE;

	return TRUE;
}

BOOL CLanCast::CreateSendingSocket(UINT nTTL, BOOL bLoopBack)
{
	if(!m_SendSocket.Create(0, SOCK_DGRAM, 0))		// Create an unconnected UDP socket
		return FALSE;

	if(!SetTTL(nTTL))								// Set Time to Live as specified by user
		AddLogLine(false,IDS_ERR_TTL);

	SetLoopBack(bLoopBack);

	return TRUE;
}

void CLanCast::Start()
{
	if (!bStarted){

		in_addr host;
		host.S_un.S_addr = GetLancastIP();
		CString t_strLanCastIP(inet_ntoa(host));
		
		// Try and join the multicast group - 224.0.0.1, using the prefs port and no loopback
		if(!JoinGroup (_T("224.0.0.1"), GetLancastPort() , 50,false))
			AddLogLine(false,IDS_JOIN_FAILED,GetLancastIP(),t_strLanCastIP,GetLancastPort());
		else {
			AddLogLine(false,IDS_JOIN_SUCCEEDED,GetLancastIP(),t_strLanCastIP,GetLancastPort());
		}
	}

}

void CLanCast::Stop()
{
	if (bStarted){
		if(!LeaveGroup())
			if (g_eMuleApp.m_pdlgEmule)	//sometimes method is called in app shutdown [TwoBottle Mod]
				AddLogLine(false,IDS_LEAVE_FAILED,GetLancastPort());
		else
			if (g_eMuleApp.m_pdlgEmule)	//sometimes method is called in app shutdown [TwoBottle Mod]
				AddLogLine(false,IDS_LEAVE_SUCCEEDED,GetLancastPort());
	}
}


BOOL CLanCast::JoinGroup(CString GroupIP, UINT nGroupPort, UINT nTTL, BOOL bLoopback)
{

	bStarted = false;

	if(!CreateReceivingSocket(GroupIP, nGroupPort))		/* Create Socket for receiving and join the host group */
		return FALSE;
	if(!CreateSendingSocket(nTTL, bLoopback))			/* Create Socket for sending */
		return FALSE;

	bStarted = true;

	return TRUE;
}

BOOL CLanCast::LeaveGroup()
{

	bStarted = false;

	if(setsockopt (m_hSocket, IPPROTO_IP, IP_DROP_MEMBERSHIP, (char FAR *)&m_mrMReq, sizeof(m_mrMReq)) < 0)
		return FALSE;

	m_SendSocket.Close();		// Close sending socket
	Close();					// Close receving socket
	return TRUE;
}

BOOL CLanCast::SetTTL(UINT nTTL)
{
	/* Set Time to Live to parameter TTL */
	if(m_SendSocket.SetSockOpt(IP_MULTICAST_TTL, &nTTL, sizeof(int), IPPROTO_IP) == 0)
		return FALSE;		/* Error Setting TTL */
	else
		return TRUE;		/* else TTL set successfully */
}

void CLanCast::SetLoopBack(BOOL bLoop)
{
	/* Set LOOPBACK option to TRUE OR FALSE according to IsLoop parameter */
	int nLoopBack = (int)bLoop;

	// Try to manually set the loopback, ie tell dozer to ignlore packets from itself (if setloopback is false)
	m_SendSocket.SetSockOpt(IP_MULTICAST_LOOP, &nLoopBack, sizeof(int), IPPROTO_IP);

}
