// EmEngine.cpp: implementation of the CEmEngine class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#ifdef NEW_SOCKETS

#include "EmWinNT.h"
#include "EmMt.h"
#include "ThreadPool.h"
#include "TaskProcessor.h"
#include "EmEngine.h"
#include "Tasks.h"
#include "../resource.h"
#include "../server.h"

//////////////////////////////////////////////////////////////////////
// start
bool CEmEngine::Init(HWND hWndUI)
{
	m_hWndUI = hWndUI;
	if (m_hStop = CreateEvent(NULL, TRUE, FALSE, NULL))
		if (m_stMain.Init())
			if (m_stTcp.Init())
			{
				// add relevant interfaces
				//m_stTcp.AddInterface(80, T_CLIENT_WEB);
				//m_stTcp.AddInterface(7821, T_CLIENT_PEER);
				return true;
			} else
			{
				ASSERT(FALSE);
				AddDebugLogLine(false, "Failed to initialize tcp engine");
			}
		else
		{
			ASSERT(FALSE);
			AddDebugLogLine(false, "Failed to initialize main engine");
		}
	else
	{
		ASSERT(FALSE);
		AddDebugLogLine(false, "CreateEvent failed");
	}

	return false;
}

//////////////////////////////////////////////////////////////////////
// stop
void CEmEngine::Uninit()
{
	if (m_hStop)
	{
		if (!SetEvent(m_hStop))
		{
			ASSERT(FALSE);
			AddDebugLogLine(false, "SetEvent failed");
		}

		m_stTcp.Uninit(false);
		m_stMain.Uninit();

		m_stTcp.Uninit(); // final

		if (!CloseHandle(m_hStop))
		{
			ASSERT(FALSE);
			AddDebugLogLine(false, "CloseHandle failed");
		}
		m_hStop = NULL;
	}
}

//////////////////////////////////////////////////////////////////////
// connecting to given server (or any server, if pServer is NULL)
void CEmEngine::ConnectToServer(CServer* pServer /* = NULL */)
{
	EMULE_TRY

	if(m_stData.GetConnectingServer())
		DisconnectFromServer(m_stData.GetConnectingServer());

	CTask_Connect *pTask = new CTask_Connect(pServer);
	m_stData.SetConnectionState(false, SERVER_CONNECTING);
	m_stMain.Push(pTask);

	EMULE_CATCH2
}

//////////////////////////////////////////////////////////////////////
// connect to any server
void CEmEngine::ConnectToAnyServer()
{
	ConnectToServer();
}

//////////////////////////////////////////////////////////////////////
// disconnect from connected server
void CEmEngine::DisconnectFromServer(SOCKET hPrevServer /* = NULL */)
{
	if(!hPrevServer)
	{
		if(m_stData.GetConnectingServer())
			DisconnectFromServer(m_stData.GetConnectingServer());
		else if(m_stData.GetConnectedServer())
			DisconnectFromServer(m_stData.GetConnectedServer());
		else
			m_stData.SetConnectionState(false, SERVER_DISCONNECTED);
		return;
	}

	CEmClient* pServer = m_stMain.Lookup(hPrevServer);
	if(!pServer)
	{
		ASSERT(FALSE);
		AddDebugLogLine(false, _T("lookup for server failed"));
		return;
	}
	CTask_KillClient* pTask = new CTask_KillClient(pServer);
	m_stMain.Push(pTask);
}

//////////////////////////////////////////////////////////////////////
// send list of shared files
void CEmEngine::SendSharedListToServer()
{
	EMULE_TRY

	CTask_SendSharedList *pTask = new CTask_SendSharedList();
	m_stMain.Push(pTask);

	EMULE_CATCH2
}


#endif