// OpCode.h: interface for the COpCode class.
//
//////////////////////////////////////////////////////////////////////
#ifdef NEW_SOCKETS
#if !defined(AFX_OPCODE_H__E3263944_435F_44B4_A12E_EAA19AAE08AD__INCLUDED_)
#define AFX_OPCODE_H__E3263944_435F_44B4_A12E_EAA19AAE08AD__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TaskProcessor.h"
#include "Tasks.h"
#include "../opcodes.h"

struct CEmClient;

const BYTE OP_PROT_EDONKEY	= 0xE3;
const BYTE OP_PROT_EMULE	= 0xC5;
const BYTE OP_PROT_PACKED	= 0xD4;
const BYTE OP_PROT_UNICODE	= 0xAA; // must be agreed

const BYTE OP_TRANSPORT_TCP = 0x00;
const BYTE OP_TRANSPORT_UDP = 0x01;

const BYTE OP_TAG_DWORD = 3;
const BYTE OP_TAG_STRING = 2;

struct CTag_Pure { // not a valid class.
	bool m_bValid;
	CTag_Pure() : m_bValid(false) {}
};

struct CTag_DWORD : public CTag_Pure {
	DWORD m_dwValue;
	CTag_DWORD() : m_dwValue(0) {}
};

struct CTag_String : public CTag_Pure {
	CString m_strValue;
};

struct CTagEntry {
	CTag_Pure*	m_pTag;
	BYTE		m_nType;
	USHORT		m_nNameLen;
	DWORD		m_dwName; // either string or number
};

struct CStream {

	bool m_bUnicode; // reserved for use with our new protocol.
	CStream() : m_bUnicode(false) {}

	class CException {};

	virtual void Write(PCVOID, DWORD);
	virtual void Read(PVOID, DWORD);
	virtual DWORD SizeLeft();
	void Skip(DWORD);

	static void Exception();

	template <class T> void operator << (const T& stObj) { Write(&stObj, sizeof(stObj)); }
	template <class T> void operator >> (T& stObj) { Read(&stObj, sizeof(stObj)); }
	
	template <> void operator << (const CString& strTxt);
	template <> void operator >> (CString& strTxt);

	void DumpString(const CString&);
	void InitString(CString&, USHORT nLen);
	void InitTrailingString(CString& strTxt);

	template <class T> void DumpArr(const CArray<T, T&>& arrValue, int nCounterLen);
	template <class T> void InitArr(CArray<T, T&>& arrValue, int nCounterLen);

	void DumpTags(CTagEntry* pTags, DWORD dwCount);
	void InitTags(CTagEntry* pTags, DWORD dwCount);

private:

	struct CQuickBuf {
		BYTE m_pBuf[0x1000]; // usually enough
		PVOID m_pBufUse;
		
		CQuickBuf() : m_pBufUse(m_pBuf) {}
		~CQuickBuf() { Free(); }

		void Alloc(DWORD dwSize)
		{
			Free();
			if ((dwSize > sizeof(m_pBuf)) &&
				!(m_pBufUse = new BYTE[dwSize]))
				CStream::Exception();

		}
		void Free()
		{
			if (m_pBufUse && (m_pBuf != m_pBufUse))
				delete m_pBufUse;
			m_pBufUse = m_pBuf;
		}
	};
};

// does not really store info. Just counts
struct CStream_Measure : public CStream {
	DWORD m_dwSize;
	virtual void Write(PCVOID, DWORD);

	inline CStream_Measure() : m_dwSize(0) {}
};

struct CStream_Mem : public CStream {
	PBYTE m_pPtr;
	DWORD m_dwSize;
	virtual void Write(PCVOID, DWORD);
	virtual void Read(PVOID, DWORD);

	inline CStream_Mem() : m_pPtr(NULL), m_dwSize(0) {}
};

struct CStream_MemEx : public CStream {
	PBYTE m_pPtr;
	DWORD m_dwSize;
	DWORD m_dwSizeExtra;
	virtual void Write(PCVOID, DWORD);

	inline CStream_MemEx() : m_pPtr(NULL), m_dwSize(0), m_dwSizeExtra(0) {}
};

struct ADDRPORT {
	ULONG	_Addr;
	USHORT	_Port;
};

class COpCode : public CTask_Tcp
{
protected:

	virtual void Z_Dump(CStream&) const = 0;
	virtual void Z_Init(CStream&) = 0;
	virtual BYTE Z_Protocol() const = 0;

#ifndef OPCODE_SKIP_DBGDUMP
	struct CDbgDump {
		CString m_strTxt;
		void WriteParam(LPCTSTR szName, DWORD dwValue);
		void WriteParam(LPCTSTR szName, const CString&);
		void WriteParam(LPCTSTR szName, const ADDRPORT&);
		void WriteParam(LPCTSTR szName, const void* pBuf, DWORD dwCount, DWORD dwElementSize);
		void WriteParam(LPCTSTR szName, const CTag_DWORD&);
		void WriteParam(LPCTSTR szName, const CTag_String&);
		template <class T> void WriteParam(LPCTSTR szName, const CArray<T, T&>&);
	};
	virtual void Z_DbgDump(CDbgDump&) const = 0;
#endif // OPCODE_SKIP_DBGDUMP

	// CTask overridables
	virtual bool Process(CEmEngine&);
	virtual bool ProcessForClient(CEmEngine&, CEmClient* pClient) = 0;

public:

	virtual ~COpCode();

	bool Write(CStream&, BYTE nTransport) const;
	static COpCode* CreateRaw(BYTE nID, BYTE nProtocol, T_CLIENT_TYPE);
	static COpCode* Read(CStream&, BYTE nID, BYTE nProtocol, T_CLIENT_TYPE eType);

	virtual BYTE GetID() const = 0;
	DWORD GetSize(BYTE nTransport) const;

#ifndef OPCODE_SKIP_DBGDUMP

	int ShowSelf(bool bMessageBox = true, int nStyle = MB_ICONINFORMATION | MB_OK | MB_TOPMOST);

#endif // OPCODE_SKIP_DBGDUMP

};


// declare opcodes
#define PARAM_DWORD(name)	PARAM_SIMPLE_ORD(name, DWORD)
#define PARAM_USHORT(name)	PARAM_SIMPLE_ORD(name, USHORT)
#define PARAM_BYTE(name)	PARAM_SIMPLE_ORD(name, BYTE)

#define PARAM_ADDRPORT(name)	PARAM_SIMPLE(name, ADDRPORT)
#define PARAM_STRING(name)		PARAM_SIMPLE(name, CString)
#define PARAM_HASH(name)		PARAM_BUF(name, BYTE, 16)

#define PARAM_INTERNAL_DEFINITION
#define PARAM_SIMPLE_ORD PARAM_SIMPLE

#ifndef OPCODE_SKIP_DBGDUMP

#	define BEGIN_OPCODE(id, name, prot, source) \
	const BYTE OP_CODE_##name = id; \
	class COpCode_##name : public COpCode { \
		virtual void Z_Dump(CStream&) const;\
		virtual void Z_Init(CStream&);\
		virtual BYTE Z_Protocol() const;\
		virtual void Z_DbgDump(CDbgDump&) const;\
		virtual bool ProcessForClient(CEmEngine&, CEmClient* pClient);\
	public:\
		virtual BYTE GetID() const;\
		COpCode_##name();\
		virtual ~COpCode_##name();\

#else // OPCODE_SKIP_DBGDUMP

#	define BEGIN_OPCODE(id, name, prot, source) \
	const BYTE OP_CODE_##name = id; \
class COpCode_##name : public COpCode { \
	virtual void Z_Dump(CStream&) const;\
	virtual void Z_Init(CStream&);\
	virtual BYTE Z_Protocol() const;\
	virtual bool ProcessForClient(CEmEngine&, CEmClient* pClient);\
public:\
	virtual BYTE GetID() const;\
	COpCode_##name();\
	virtual ~COpCode_##name();\

#endif // OPCODE_SKIP_DBGDUMP

#define END_OPCODE };

#define OPCODE_PROCESS_UI virtual bool ProcessForUI(CEmEngine&);

#define PARAM_BYTE_JUNK(value)
#define PARAM_BUF(name, type, count) type _##name[count];
#define PARAM_SIMPLE(name, type) type _##name;
#define PARAM_TAGS_BEGIN
#define PARAM_TAGS_END
#define PARAM_TAG_STR(name, id) CTag_String _##name;
#define PARAM_TAG_DWORD(name, id) CTag_DWORD _##name;
#define PARAM_NTAG_DWORD(name, id) CTag_DWORD _##name;
#define PARAM_TEXT_TRAILING(name) CString _##name;
#define PARAM_ARRAY(name, type, counter) CArray<type, type&> _##name;

#ifndef OPCODE_SKIP_DBGDUMP
#define PARAM_COMPLEXARRAY_BEGIN(msg, name) struct CStruct_##name { CStruct_##name(); void Z_Dump(CStream&) const; void Z_Init(CStream&); void Z_DbgDump(CDbgDump&) const;
#else
#define PARAM_COMPLEXARRAY_BEGIN(msg, name) struct CStruct_##name { CStruct_##name(); void Z_Dump(CStream&) const; void Z_Init(CStream&);
#endif // OPCODE_SKIP_DBGDUMP

#define PARAM_COMPLEXARRAY_END(name) }; CArray<CStruct_##name, CStruct_##name&> _##name;

#include "EmMsgs.h"

#undef BEGIN_OPCODE
#undef END_OPCODE

#undef OPCODE_PROCESS_UI
#define OPCODE_PROCESS_UI

#undef PARAM_BYTE_JUNK
#undef PARAM_BUF
#undef PARAM_SIMPLE
#undef PARAM_TAGS_BEGIN
#undef PARAM_TAGS_END
#undef PARAM_TAG_STR
#undef PARAM_TAG_DWORD
#undef PARAM_NTAG_DWORD
#undef PARAM_TEXT_TRAILING
#undef PARAM_ARRAY
#undef PARAM_INTERNAL_DEFINITION
#undef PARAM_COMPLEXARRAY_BEGIN
#undef PARAM_COMPLEXARRAY_END


#endif // !defined(AFX_OPCODE_H__E3263944_435F_44B4_A12E_EAA19AAE08AD__INCLUDED_)
#endif