//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "zlib/zlib.h"
#include "UpDownClient.h"
#include "opcodes.h"
#include "packets.h"
#include "emule.h"
#include "UploadQueue.h"
#include "otherstructs.h"
#include "otherfunctions.h"
#include "math.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif


//	members of CUpDownClient
//	which are mainly used for uploading functions

void CUpDownClient::DrawUpStatusBar(CDC* dc, RECT* rect, bool onlygreyrect, bool bFlat)
{
	EMULE_TRY

	const COLORREF crBoth = (bFlat) ? RGB(0, 150, 0) : RGB(0, 192, 0);	//Cax2 - same green used for files & clients
	const COLORREF crNeither = (bFlat) ? RGB(224, 224, 224) : RGB(240, 240, 240);
	const COLORREF crBeingSent = RGB(255, 208, 0);

	uint32 filesize = (m_uUpPartCount) ? PARTSIZE * (m_uUpPartCount) : 1; // might be a little off.. better than nothing

	CBarShader statusBar(rect->bottom - rect->top, rect->right - rect->left); //Maella - Not a static object anymore
	statusBar.SetFileSize(filesize);
	statusBar.Fill(crNeither);

	if (filesize != 1 && !onlygreyrect && m_pbyteUpPartStatuses)	//Cax2 - no more black rectangle
	{
		for (uint32 i = 0;i != m_uUpPartCount;i++)
		{
			/* no need for that now, cause we don't use the precise filesize
			if (PARTSIZE*(i+1) > filesize) 
			uEnd = filesize; 
			else 
			uEnd = PARTSIZE*(i+1); 
			*/
			uint32 uEnd = PARTSIZE * (i + 1);

			if (m_pbyteUpPartStatuses[i])
			{ //they have this part
				statusBar.FillRange(PARTSIZE * i, uEnd, crBoth); // will do for now
			}
			else //they don't have a part.
				statusBar.FillRange(PARTSIZE * i, uEnd, crNeither);
			uint32 partno = GetCurrentlyUploadingPart();
			if (i == partno) // won't be equal if it is 0xFFFF
				statusBar.FillRange(PARTSIZE * i, uEnd, crBeingSent);
		}
	}
	statusBar.Draw(dc, rect->left, rect->top, bFlat);

	EMULE_CATCH
}

bool CUpDownClient::IsCommunity()
{
	EMULE_TRY

	if (!m_bCommunityIsCached)
	{
	//	[i_a] DonGato (Community Sharing)
	//eklmn: to prevent a leeching the community string must have a same format
		CString username(m_strUserName);
		CString community(g_eMuleApp.m_pGlobPrefs->CommunityString());
		CString communityok;

		community.Remove( '[' );
		community.Remove(']');
		communityok.Format("[%s]", community);
		m_bIsCommunity = (username.Find(communityok) >= 0);
		m_bCommunityIsCached = true;
	}
	return m_bIsCommunity;

	EMULE_CATCH
	return false;
}

void CUpDownClient::ResetValueCache()
{
	EMULE_TRY
	m_bScoreBaseIsCached = false;
	m_bCommunityIsCached = false;
	EMULE_CATCH
}

uint32 CUpDownClient::GetScore(bool sysvalue, bool isdownloading, bool onlybasevalue)
{
	static DWORD	dwLastMsgTime = 0;	//spam control

	EMULE_TRY
//	TODO: complete this (friends, uploadspeed, emuleuser etc etc)
	if (m_strUserName.IsEmpty())
		return 0;
//SyruS moved this up (from end) to skip the whole useless calculation
//Start milobac: Anti-leechermods
	if (m_bBanned)
		return 0;
	else
		if (IsStolenName() || IsLeecherType() /*&& g_eMuleApp.m_pGlobPrefs->GetMlDonkeyCM()*/)
		{
			if (::GetTickCount() > dwLastMsgTime + (60 * 1000))	//spam control
			{
				if (!m_strModVersion.IsEmpty())
					AddDebugLogLine(false, _T("%s: Anti-leechermods: Client %s (%s) using the mod %s has been banned"), __FUNCTION__, GetUserName(), GetClientNameAndVersionString(this), GetModVersion());
				else
					AddDebugLogLine(false, _T("%s: Anti-leechermods: Client %s (%s) has been banned"), __FUNCTION__, GetUserName(), GetClientNameAndVersionString(this));
			//katsyonak: Send leecher ban message 15.06.03
				Ban(_T("AUTOMATED MESSAGE: You're being banned because your client is leeching!"), GetResString(IDS_BAN_LEECHERADV));

				dwLastMsgTime = ::GetTickCount();
			}
			return 0;
		}
//End milobac: Anti-leechermods

#ifdef OLD_SOCKETS_ENABLED
	if (!isdownloading && !onlybasevalue)
	{
		if (HasLowID() && !(m_pRequestSocket && m_pRequestSocket->IsConnected()))
		{
			if (!g_eMuleApp.m_pServerConnect->IsConnected() || g_eMuleApp.m_pServerConnect->IsLowID())
				return 0;
		}
	}
#endif //OLD_SOCKETS_ENABLED

	CKnownFile* req_file = g_eMuleApp.m_pSharedFilesList->GetFileByID(m_reqFileHash);
	if (req_file == NULL)
		return 0;

//eklmn: we cann't cache a value here, cause otherwise our chages of priority will
//	be not applied until we restart eMule.
	ASSERT(m_reqFileHash != NULL); // i_a: m_reqFileHash is an array so the if() was superfluous
	{
		switch (req_file->GetPriority())
		{
			case PR_VERYHIGH:  // Modified by Tarod for UAP
			if (req_file->IsPartFile())
				m_uFilePriority = 50;
			else
				m_uFilePriority = 100;
			break;

			case PR_HIGH:
			m_uFilePriority = 20;
			break;

			case PR_LOW:
			m_uFilePriority = 5;
			break;

			case PR_VERYLOW:
			m_uFilePriority = 2;
			break;

			case PR_NORMAL:
			default:
			m_uFilePriority = 10;	// standard
			break;
		}
	}
//End milobac: Anti-leechermods

//	calculate score, based on waiting time and other factors
	double		dblBaseValue;
	if (onlybasevalue)
		dblBaseValue = 100.0f;
	else if (!isdownloading)
		dblBaseValue = static_cast<double>(GetTickCount() - m_dwWaitTime) / 1000.0;
	else
	{
	//	we don't want one client to download forever
	//	the first 15 min download time counts as 15 min waitingtime and you get a 15 min bonus while you are in the first 15 min :)
	//	(to avoid 20 sec downloads) after this the score won't raise anymore
		dblBaseValue = static_cast<double>(m_dwUploadTime - m_dwWaitTime);
		ASSERT (m_dwUploadTime - m_dwWaitTime >= 0); //oct 28, 02: changed this from "> 0" to ">= 0"
		dblBaseValue += static_cast<double>((GetTickCount() - m_dwUploadTime > 900000) ? 900000.0 : 1800000.0);
		dblBaseValue /= 1000.0;
	}
	if (GetDownloadState() == DS_DOWNLOADING && GetDataRate() > 500) // obaldin: prevent 0.1 upload abuse
		dblBaseValue *= 1.5; // since up/down credits are in, this isn't that important anymore
	if (IsFriend())
		dblBaseValue *= 2.0;
//	DonGato (Community Sharing)
	if (g_eMuleApp.m_pGlobPrefs->CommunityEnabled() && IsCommunity())
		dblBaseValue *= 2.0;

//Fix by Amdribant [loadedsrc]
//ASSERT(m_pCredits != NULL);
	if (m_pCredits != (CClientCredits*)NULL)
	{
		dblBaseValue *= m_pCredits->GetScoreRatio(this->GetIP());
	}

//	Added by Tarod (herbert code)
	dblBaseValue *= GetSmallFilePushRatio();
	dblBaseValue *= GetRareFilePushRatio();

	dblBaseValue *= static_cast<double>(m_uFilePriority) / 10.0;

	return static_cast<uint32>(dblBaseValue);

	EMULE_CATCH
	return 1;
}

//	Added by Tarod based in herbert code and modifications by MatzeHH
double CUpDownClient::GetRareFilePushRatio()
{
	EMULE_TRY
//if(!g_eMuleApp.m_pGlobPrefs->FairShareRarePush())
//	return 1.0;
	CKnownFile * srcfile = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)m_reqFileHash);
	if (srcfile == NULL)
		return 1.0;

//	keep the FileRatio
	double ratio = 0 + srcfile->GetFileRatio();
	return(ratio < 1.0 ? 1.0 : ((ratio > 100.0) ? 100.0 : ratio)); //Cax2 bugfix

	EMULE_CATCH
	return 1.0;
}

//	Checks if it is next requested block from another chunk of the actual file or from another file
//
//	[Returns]
//	true : Next requested block is from another different chunk or file than last downloaded block
//	false: Next requested block is from same chunk that last downloaded block
bool CUpDownClient::IsDifferentPartBlock(bool silent) // [Tarod 12/22/2002] Cax2 - added silent option
{
	Requested_Block_Struct * last_done_block;
	Requested_Block_Struct* next_requested_block;
	uint32 last_done_part = 0xffffffff;
	uint32 next_requested_part = 0xffffffff;

	bool different_part = false;

	try
	{
	//	Check if we have good lists and proceed to check for different chunks
		if (!m_blockRequestsQueue.IsEmpty() && !m_doneBlocksList.IsEmpty())
		{
		//	Get last block and next pending
			last_done_block = (Requested_Block_Struct*)m_doneBlocksList.GetHead();
			next_requested_block = (Requested_Block_Struct*)m_blockRequestsQueue.GetHead();

		//	Calculate corresponding parts to blocks
			last_done_part = last_done_block->m_dwStartOffset / PARTSIZE;
			next_requested_part = next_requested_block->m_dwStartOffset / PARTSIZE;

		//	Test is we are asking same file and same part
			/*
			if ( last_done_part != next_requested_part)
			{ 
				different_part = true;
				//AddDebugLogLine(false, "Session for user %s (%s) ended due to new chunk, sent %u KB", GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024); 
			}
			if (md4cmp (last_done_block->m_fileHash, next_requested_block->m_fileHash) != 0)
			{ 
				different_part = true;
			             //AddDebugLogLine(false, "Session for user %s(%s) ended due to different file, sent %u KB", GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024); 
			}
			*/
			if (last_done_part != next_requested_part || md4cmp(last_done_block->m_fileHash, next_requested_block->m_fileHash) != 0)
				different_part = true;
		}
	}
	catch (...)
	{
		different_part = true;
	}

	if (different_part)
	{
	//AddDebugLogLine(false, "Debug: User %s, last_done_part (%u) %s (%u) next_requested_part, sent %u KB", GetUserName(), last_done_part, different_part? "!=": "==", next_requested_part, this->GetTransferredUp() / 1024);

	//	obaldin: this is for old clients, that rotate the chunk,
	//	so that they'll be able to download at least _something_
	//	But probably even after the majority switches we'd
	//	better check for 0 here
		if (GetSessionUp() < 1024 * 1024)
		{
			different_part = false;
			if (!silent)
				AddDebugLogLine(false, _T("Client %s continues downloading (should be back on queue) because of a low transferred size"), GetUserName());
		}
	}

	return different_part;
}

#if JUMPSTART
bool CUpDownClient::IsForbiddenByJumpstart()
{
	EMULE_TRY

	Requested_Block_Struct * next_requested_block;
	uint32 next_requested_part = 0xffffffff;

	if (m_blockRequestsQueue.IsEmpty())
	{
		return false;
	}
	else
	{
		next_requested_block = (Requested_Block_Struct*)m_blockRequestsQueue.GetHead();
		next_requested_part = next_requested_block->m_dwStartOffset / PARTSIZE;
		CKnownFile* kf = g_eMuleApp.m_pSharedFilesList->GetFileByID(next_requested_block->m_fileHash);
		if (kf != NULL && !kf->AllowChunkForClient(next_requested_part, this))
		{
			return true;
		}
	}

	EMULE_CATCH

	return false;
}
#endif // JUMPSTART


double CUpDownClient::GetSmallFilePushRatio()
{
	EMULE_TRY
//if(!g_eMuleApp.m_pGlobPrefs->FairShareRarePush())
//	return 1.0f;
	CKnownFile * srcfile = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)m_reqFileHash);
	if (srcfile == NULL)
		return 1.0;

//Cax2 faster code ....
	long lFileSize = srcfile->GetFileSize() ;

//if (filesizeinmb > 0.0f && filesizeinmb < 9.5f) {
	if (lFileSize > 0.0 && lFileSize < 9961472)
	{

	//	Modded by Tarod to: Min(100, Max(1, 9.5 / FS)) [9.5 Mbs. == Chunk Size]
		double		mul = 9961472.0 / lFileSize ;		// VQB formula + Cax2 overflow bugfix...

		if (mul > 100.0)
			mul = 100.0 ;		// VQB formula
		return mul ;
	}

	EMULE_CATCH
	return 1.0;
}
//	End added by Tarod

//SyruS (0.29c)
class CSyncHelper
{
public:

	CSyncHelper()
	{
		m_pObject = NULL;
	}

	~CSyncHelper()
	{
		if (m_pObject)
			m_pObject->Unlock();
	}

	CSyncObject* m_pObject;
};

bool CUpDownClient::CreateNextBlockPackage()
{
//	check if we should kick this client (time critical !!!)
//	Never kick friends
//	Always kick if transfer is longer than 80 minutes
//	Never allow more than the size of a chunk (+500 KBytes for small resends of invalid data) to be transferred
//	VQB to provide full chunk transfers (modified by Tarod)
//	Note: Full chunk transfers really should be enforced, not an option
	EMULE_TRY
//	LANCAST (moosetea) We never kick a lancast user
	bool AllowedMore = IsOnLAN();
//	v- eklmn: bugfix(04): conflict between upload logic & LANcast
	bool diff_part = IsDifferentPartBlock(true);	//Cax2 - don't write anyting to the log yet!
#if JUMPSTART

	bool JS_blocked = IsForbiddenByJumpstart();
#endif

	bool shouldGo = ((GetUpStartTimeDelay() >= 80 * 60 * 1000) || diff_part);

//	if user is from Internet && full chunk (or timeout)
	if (!AllowedMore && shouldGo)
	{
#if JUMPSTART
		if (IsFriend() && !g_eMuleApp.m_pUploadQueue->CheckForTimeOver(this) && !JS_blocked)
#else

		if (IsFriend() && !g_eMuleApp.m_pUploadQueue->CheckForTimeOver(this))
#endif

		{
			AllowedMore = true;
			AddDebugLogLine(false, _T("Client %s downloaded %u KB so far. Allowed to continue downloading (should go back on queue) because it's a friend."), GetUserName(), this->GetSessionUp() / 1024);
		}
		else
		{
			SetWaitStartTime();
			if (diff_part)
			{
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
				AddDebugLogLine(false, _T("Client %s (%s)  back on queue. Asked for a new chunk/file, after receiving %u KB."), GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024);
			}
#if JUMPSTART
			else if (JS_blocked)
			{
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
				AddDebugLogLine(false, _T("Client %s (%s)  back on queue. Asked for a chunk forbidden by Jumpstart, after receiving %u KB."), GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024);
			}
#endif
			else
			{
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_TIMEOUT);
				AddDebugLogLine(false, _T("Client %s (%s)  back on queue. Upload timeout, after receiving %u KB"), GetUserName(), GetClientNameAndVersionString(this), GetSessionUp() / 1024);
			}
			g_eMuleApp.m_pUploadQueue->AddClientToQueue(this, true);
			return false;
		}
	}
//	^- eklmn: bugfix(04): conflict between upload logic & LANcast

	if (m_blockRequestsQueue.IsEmpty())
	{
	//	There are no new blocks requested, put we will happily transfer more
	//	if the client just requests something
	//	AddDebugLogLine(false, "No more data to send to client %s(%s), should be back on queue", GetUserName(), GetClientNameAndVersionString(this));
		return false; //eklmn: bugfix(03): keeping user in queue if he doesn't need anything
	}

	CFile file;
	CKnownFile* srcfile;
	byte* filedata = 0;
	bool bFromPF = true;
	CString fullname;
	CSyncHelper lockFile;
	try
	{
	//	repeat as long as next requested block is in the same chunk as previous blocks
	//	and as long as there is a next requested block
	//added: if a friend or lancast user, keep going![Cax2]
#if JUMPSTART
		while (!m_blockRequestsQueue.IsEmpty() && (AllowedMore || !IsDifferentPartBlock()) && !IsForbiddenByJumpstart())
#else

		while (!m_blockRequestsQueue.IsEmpty() && (AllowedMore || !IsDifferentPartBlock()))		//Cax2 - if a friend or lancast user, keep going!
#endif

		{
			Requested_Block_Struct * currentblock = m_blockRequestsQueue.GetHead();
			srcfile = g_eMuleApp.m_pSharedFilesList->GetFileByID(currentblock->m_fileHash);
			if (!srcfile)
			{
				g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);	// eklmn: ???
				throw GetResString(IDS_ERR_REQ_FNF);
			}

			if (srcfile->IsPartFile() && ((CPartFile*)srcfile)->GetStatus() != PS_COMPLETE)
			{
			//	SyruS (0.29c) Do not access a part file, if it is currently moved into the incoming directory.
			//	Because the moving of part file into the incoming directory may take a noticable
			//	amount of time, we can not wait for 'm_FileCompleteMutex' and block the main thread.
				if (!((CPartFile*)srcfile)->m_FileCompleteMutex.Lock(0)) // just do a quick test of the mutex's state and return if it's locked.
					return false;
				lockFile.m_pObject = &((CPartFile*)srcfile)->m_FileCompleteMutex;
			//	If it's a part file which we are uploading the file remains locked until we've read the
			//	current block. This way the file completion thread can not (try to) "move" the file into
			//	the incoming directory.
				fullname = ((CPartFile *)srcfile)->GetFullName();
				fullname.Truncate(fullname.GetLength() - 4);
			}
			else
			{
				fullname.Format(_T("%s\\%s"), srcfile->GetPath(), srcfile->GetFileName());
			}

		//v- eklmn: bugfix(18): conflict between JUMSTART & PartTraffic)
#if JUMPSTART
			if (!srcfile->GetJumpstartEnabled())
#else

			if (true)
#endif

			{
			//--- xrmb:partprio ---
			//--- accessing a blocked part??? ---
				uint8 part_status = srcfile->GetPartStatus(currentblock->m_dwStartOffset / PARTSIZE);
				CString s;
				switch (part_status)
				{
					case PR_PART_OFF:
					AddLogLine(false, IDS_PARTPRIO_DENIED, currentblock->m_dwStartOffset / PARTSIZE, srcfile->GetFileName());
					s.Format(GetResString(IDS_PARTPRIO_DENIED), currentblock->m_dwStartOffset / PARTSIZE, srcfile->GetFileName());
					g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
					throw s;
					break;
					case PR_PART_HIDDEN:
					AddLogLine(false, IDS_PARTPRIO_HIDDEN, GetUserName(), srcfile->GetFileName(), currentblock->m_dwStartOffset / PARTSIZE);
				}
			//--- :xrmb ---
			}
		//^- eklmn: bugfix(18): conflict between JUMSTART & PartTraffic)

			uint32 togo;
			if (AllowedMore)
			{	// Friends & LAN  are not limited by chunks boundaries.
			//	Is this a wraparound request?
				if (currentblock->m_dwStartOffset > currentblock->m_dwEndOffset - 1)
				{
				//	If m_dwStartOffset is later in the file than m_dwEndOffset, it means
				//	that the request is a wraparound request. It reads to the end of
				//	the file, then it wraps around and continues in the beginning of the file.

				//	Calculate how much the client is requesting, knowing that this is a wraparound request
					togo = currentblock->m_dwEndOffset + (srcfile->GetFileSize() - currentblock->m_dwStartOffset);
				}
				else
				{
				//	not a wraparound request. This makes it easier to calculate how many bytes is
				//	being requested.
					togo = currentblock->m_dwEndOffset - currentblock->m_dwStartOffset;
				}
			}
			else
			{
			//	Everyone else are limited to a single chunk
			//	Check that m_dwStartOffset and m_dwEndOffset is in the same chunk
				if (currentblock->m_dwStartOffset / PARTSIZE != (currentblock->m_dwEndOffset - 1) / PARTSIZE)
				{
				//	The m_dwEndOffset goes into the next chunk. Set it to the end of the chunk
				//	that m_dwStartOffset is in.
					currentblock->m_dwEndOffset = ((uint32)(currentblock->m_dwStartOffset / PARTSIZE) + 1) * PARTSIZE;
				}

			//	Make sure we don't pass end of file
				if (currentblock->m_dwStartOffset > currentblock->m_dwEndOffset)
				{
					currentblock->m_dwEndOffset = srcfile->GetFileSize();
				}

			//	This can't be a wrapped around request, since it has been limited to
			//	a single chunk.
				togo = currentblock->m_dwEndOffset - currentblock->m_dwStartOffset;
			}

			if (srcfile->IsPartFile())
			{ // proper check
				uint32 parts_start = ((currentblock->m_dwStartOffset) / PARTSIZE) * PARTSIZE;
				uint32 parts_end = ((currentblock->m_dwEndOffset - 1) / PARTSIZE + 1) * PARTSIZE - 1;
				if (!((CPartFile*)srcfile)->IsComplete(parts_start, parts_end))
				{
					g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
					throw GetResString(IDS_ERR_INCOMPLETEBLOCK);
				}
			}

			if (!srcfile->IsPartFile())
			{
				bFromPF = false;	// <-khaos--+++> This is not a part file...
				if (!file.Open(fullname, CFile::modeRead | CFile::osSequentialScan | CFile::shareDenyNone))
				{
					g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);
					throw GetResString(IDS_ERR_OPEN);
				}
				file.Seek(currentblock->m_dwStartOffset, 0);

				filedata = new byte[togo + 500];
				if (uint32 done = file.Read(filedata, togo) != togo)
				{
					file.SeekToBegin();
					file.Read(filedata + done, togo - done);
				}
				file.Close();
			}
			else
			{
				CPartFile* partfile = (CPartFile*)srcfile;
				partfile->GetPartFileHandle().Seek(currentblock->m_dwStartOffset, 0);

				filedata = new byte[togo + 500];
				if (uint32 done = partfile->GetPartFileHandle().Read(filedata, togo) != togo)
				{
					partfile->GetPartFileHandle().SeekToBegin();
					partfile->GetPartFileHandle().Read(filedata + done, togo - done);
				}
			}
			if (lockFile.m_pObject)
			{
				lockFile.m_pObject->Unlock();	//SyruS (0.29c) Unlock the (part) file as soon as we are done with accessing it.
				lockFile.m_pObject = NULL;		//SyruS (0.30b)
			}

			md4cpy(m_reqFileHash, currentblock->m_fileHash);
			CString ext = srcfile->GetFileName().Right(4);
		//	-khaos--+++> We're going to add bFromPF as a parameter to the calls to create packets...
			if (m_byteDataCompVer == 1 && ext != _T(".zip") && ext != _T(".rar") && ext != _T(".ace"))
				CreatePackedPackets(filedata, togo, currentblock, bFromPF);
			else
				CreateStandardPackets(filedata, togo, currentblock, bFromPF);
		//	<-----khaos-

		//	file statistic
			srcfile->statistic.AddTransferred(togo);

		//--- xrmb:parttraffic ---
			srcfile->statistic.AddTraffic(currentblock->m_dwStartOffset, togo);
		//--- :xrmb ---

#if JUMPSTART

			if (!srcfile->IsPartFile())
			{
				if (srcfile->GetJumpstartEnabled())
				{
					srcfile->AddSentBlock(this, currentblock->m_dwStartOffset, togo);
				}
			}
#endif // JUMPSTART

			m_doneBlocksList.AddHead(m_blockRequestsQueue.RemoveHead());
			delete[] filedata;
			filedata = 0;
		}
	}
	catch (CString & error)
	{
		OUTPUT_DEBUG_TRACE();
		AddDebugLogLine(false, IDS_ERR_CLIENTERRORED, GetUserName(), error.GetBuffer());
		if (filedata)
			delete filedata;
		return false;
	}
//v- eklmn: bugfix(13): memleak in UploadClient due CFile (based on 0.29a)
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
		AddDebugLogLine(false, _T("Failed to create upload package for %s - %s"), GetUserName(), GetErrorMessage(error));
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);
		if (filedata)
			delete filedata;
		error->Delete();
		return false;
	}
//^- eklmn: bugfix(13): memleak in UploadClient due CFile (based on 0.29a)
//	AddDebugLogLine(false,"Debug: Packet done. Size: %i",blockpack->GetLength());
	return true;

	EMULE_CATCH
	return false;
}

void CUpDownClient::CheckForUpPartsInfo()
{
	EMULE_TRY
	if (m_uPartCount == 0)
		return;
	if (!m_pReqPartFile)
		return;
	if (memcmp(m_reqFileHash, m_pReqPartFile->GetFileHash(), 16))
		return;
	if (!m_pbytePartStatuses)
		return;

	if (m_pbyteUpPartStatuses)
	{
		delete[] m_pbyteUpPartStatuses;
		m_pbyteUpPartStatuses = NULL;
	}
	m_uUpPartCount = m_uPartCount;
	m_uAvailUpPartCount = 0;
	m_pbyteUpPartStatuses = new uint8[m_uUpPartCount];
	for (int i = 0; i < m_uUpPartCount; i++)
	{
		m_pbyteUpPartStatuses[i] = m_pbytePartStatuses[i];
		if (m_pbyteUpPartStatuses[i])
			m_uAvailUpPartCount++;
	}
	EMULE_CATCH
}

void CUpDownClient::ProcessUpFileStatus(char* packet, uint32 size)
{
	EMULE_TRY

	if (m_pbyteUpPartStatuses)
	{
		delete[] m_pbyteUpPartStatuses;
		m_pbyteUpPartStatuses = NULL;	// added by jicxicmic
	}
	m_uUpPartCount = 0;
	m_uAvailUpPartCount = 0;
	m_uUpCompleteSourcesCount = 0;	// netwolf: complete sources (zegzav )
	if (size == 16)
	{
		CheckForUpPartsInfo(); // obaldin
		return;
	}
	CSafeMemFile data((BYTE*)packet, size);
	uchar cfilehash[16];
	data.Read(cfilehash, 16);
	CKnownFile* tempreqfile = g_eMuleApp.m_pSharedFilesList->GetFileByID(cfilehash);
	data.Read(&m_uUpPartCount, 2);
	if (!m_uUpPartCount)
	{
		m_uUpPartCount = tempreqfile->GetPartCount();
		m_pbyteUpPartStatuses = new uint8[m_uUpPartCount];
		memset2(m_pbyteUpPartStatuses, 0, m_uUpPartCount);
	}
	else
	{
		if (tempreqfile->GetPartCount() != m_uUpPartCount)
		{
			m_uUpPartCount = 0;
			return;
		}
		m_pbyteUpPartStatuses = new uint8[m_uUpPartCount];
		uint16 done = 0;
		while (done != m_uUpPartCount)
		{
			uint8 toread;
			data.Read(&toread, 1);
			for (sint32 i = 0;i != 8;i++)
			{
				m_pbyteUpPartStatuses[done] = ((toread >> i) & 1) ? 1 : 0;
				if (m_pbyteUpPartStatuses[done])
					m_uAvailUpPartCount++;
			//	We may want to use this for another feature..
			//	if (m_pbyteUpPartStatuses[done] && !tempreqfile->IsComplete(done*PARTSIZE,((done+1)*PARTSIZE)-1))
			//	bPartsNeeded = true;
				done++;
				if (done == m_uUpPartCount)
					break;
			}
		}

	//	netwolf: complete sources (zegzav ) - BEGIN
		if ((GetExtendedRequestsVersion() > 1) && (size >= 20 + (((uint32)m_uUpPartCount + 7) >> 3)))
		{
			uint16 nCount;
			data.Read(&nCount, 2);
			SetUpCompleteSourcesCount(nCount);
		}
	//	netwolf: complete sources (zegzav ) - END
	}
	tempreqfile->CalculateCompleteSources(); // netwolf complete sources (zegzav) // tempreqfile->NewAvailPartsInfo()
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}


void CUpDownClient::CreateStandardPackets(byte* data, uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF)
{
	EMULE_TRY
	uint32 nPacketSize;
	CMemFile memfile((BYTE*)data, togo);
	if (togo > 10240)
		nPacketSize = togo / (uint32)(togo / 10240);
	else
		nPacketSize = togo;
	while (togo)
	{
		if (togo < nPacketSize * 2)
			nPacketSize = togo;
		togo -= nPacketSize;
		Packet* packet = new Packet(OP_SENDINGPART, nPacketSize + 24, OP_EDONKEYPROT, bFromPF);	// -khaos: Create the packet with the new boolean.
		md4cpy(&packet->m_pcBuffer[0], m_reqFileHash);
		uint32 statpos = (currentblock->m_dwEndOffset - togo) - nPacketSize;
		uint32 endpos = (currentblock->m_dwEndOffset - togo);
		memcpy2(&packet->m_pcBuffer[16], &statpos, 4);
		memcpy2(&packet->m_pcBuffer[20], &endpos, 4);
		memfile.Read(&packet->m_pcBuffer[24], nPacketSize);

		SplitAndAddPacketsToSendQueue(packet, bFromPF);
	}
	EMULE_CATCH
}

void CUpDownClient::CreatePackedPackets(byte* data, uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF)
{
	EMULE_TRY
	BYTE * output = new BYTE[togo + 300];
	uLongf newsize = togo + 300;
	uint16 result = compress2(output, &newsize, data, togo, 9);
	if (result != Z_OK || togo <= newsize)
	{
		delete[] output;
		CreateStandardPackets(data, togo, currentblock, bFromPF);		// -khaos: Create the packet with the new boolean.
		return;
	}
	m_bUsedCompressionUp = true;
	m_dwCompressionGain += (togo - newsize); // Add show compression
	m_dwUncompressed += togo; // Add show compression
	CMemFile memfile(output, newsize);
	togo = newsize;
	uint32 nPacketSize;
	if (togo > 10240)
		nPacketSize = togo / (uint32)(togo / 10240);
	else
		nPacketSize = togo;
	while (togo)
	{
		if (togo < nPacketSize * 2)
			nPacketSize = togo;
		togo -= nPacketSize;
		Packet* packet = new Packet(OP_COMPRESSEDPART, nPacketSize + 24, OP_EMULEPROT, bFromPF);	// -khaos: Create the packet with the new boolean.
		md4cpy(&packet->m_pcBuffer[0], m_reqFileHash);
		uint32 statpos = currentblock->m_dwStartOffset;
		memcpy2(&packet->m_pcBuffer[16], &statpos, 4);
		memcpy2(&packet->m_pcBuffer[20], &newsize, 4);
		memfile.Read(&packet->m_pcBuffer[24], nPacketSize);

		SplitAndAddPacketsToSendQueue(packet, bFromPF);
	}
	delete[] output;
	EMULE_CATCH
}

/**
 * A packet should never be larger than MAXFRAGSIZE. This method splits up the
 * given packets into several packets with the size MAXFRAGSIZE, if necessary.
 */
void CUpDownClient::SplitAndAddPacketsToSendQueue(Packet* packet, bool bFromPF)
{
	EMULE_TRY

	if (packet->GetRealPacketSize() > 0 && packet->GetRealPacketSize() > MAXFRAGSIZE * 2)
	//	splitting packets
	{
		uint32 nSize = packet->GetRealPacketSize();
		char* m_pcBuffer = packet->DetachPacket();
		delete packet;

		uint32 nPos = 0;
		bool bLast = true;
		while (nPos < nSize)
		{
			uint32 nNewSize = (nSize - nPos < MAXFRAGSIZE) ? nSize - nPos : MAXFRAGSIZE;
			char* pBuffer2 = new char[nNewSize];
			memcpy2(pBuffer2, m_pcBuffer + nPos, nNewSize);
			nPos += nNewSize;
			m_blockSendQueue.AddTail(new Packet(pBuffer2, nNewSize, (nNewSize < MAXFRAGSIZE), bFromPF));
			bLast = false;
		}
		delete[] m_pcBuffer;
	}
	else
	{
		m_blockSendQueue.AddTail(packet);
	}

	EMULE_CATCH
}

void CUpDownClient::SetUploadFileID(uchar* tempreqfileid)
{
	EMULE_TRY
	if (tempreqfileid)
		md4cpy(m_reqFileHash, tempreqfileid);
	else
		for (int i = 0; i < 16; i++)
			m_reqFileHash[i] = 0;
	EMULE_CATCH
}

void CUpDownClient::AddReqBlock(Requested_Block_Struct* reqblock)
{
	EMULE_TRY
	for (POSITION pos = m_doneBlocksList.GetHeadPosition();pos != 0;m_doneBlocksList.GetNext(pos))
	{
		if (reqblock->m_dwStartOffset == m_doneBlocksList.GetAt(pos)->m_dwStartOffset && reqblock->m_dwEndOffset == m_doneBlocksList.GetAt(pos)->m_dwEndOffset)
		{
			delete reqblock;
			return;
		}
	}
	for (POSITION pos = m_blockRequestsQueue.GetHeadPosition();pos != 0;m_blockRequestsQueue.GetNext(pos))
	{
		if (reqblock->m_dwStartOffset == m_blockRequestsQueue.GetAt(pos)->m_dwStartOffset && reqblock->m_dwEndOffset == m_blockRequestsQueue.GetAt(pos)->m_dwEndOffset)
		{
			delete reqblock;
			return;
		}
	}
	m_blockRequestsQueue.AddTail(reqblock);

	EMULE_CATCH
}

void CUpDownClient::SetUpStartTime(uint32 dwTime)
{
	EMULE_TRY
	if (dwTime)
		m_dwUploadTime = dwTime;
	else
		m_dwUploadTime = ::GetTickCount();
	EMULE_CATCH
}

void CUpDownClient::SetWaitStartTime(uint32 dwTime)
{
	EMULE_TRY
	if (dwTime)
		m_dwWaitTime = dwTime;
	else
		m_dwWaitTime = ::GetTickCount();
	EMULE_CATCH
}

uint32 CUpDownClient::SendBlockData(uint32 nMaxAmmount, bool& wantsMoreBandwidth)
{
	EMULE_TRY
	DWORD curTick = GetTickCount();
//	this will be set to true if/when the slot is actually fed some data
	wantsMoreBandwidth = false;
	uint32 amounttransferred = 0;

//	check for 'stuck' clients
//	if ((GetSessionUp()==0) && (GetUpStartTimeDelay()>2*60*1000)) {
	if ((curTick - GetLastGotULData() > 2 * 60 * 1000) && (GetUpStartTimeDelay() > 2 * 60 * 1000))
	{
	//	were unable to transfer for 2 minutes, put the client back on queue
		AddDebugLogLine(false, _T("Client %s(%s) appears to be stuck, putting back on queue"), GetUserName(), GetClientNameAndVersionString(this));
		SetWaitStartTime();
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_TIMEOUT);
		g_eMuleApp.m_pUploadQueue->AddClientToQueue(this, true);
		return 0;
	}

#ifdef OLD_SOCKETS_ENABLED
//	Only feed this slot new data if it has been able to empty its
//	buffer since we last offered it data
	if (m_pRequestSocket && !m_pRequestSocket->IsBusy() && GetUploadState() == US_UPLOADING)
	{
	//	v- eklmn: bugfix(03): keeping user in queue if he doesn't need anything (v2)
		if (m_blockSendQueue.IsEmpty())
		{
		//	no valid blocks to transfer. Put the client back on queue
			if (!CreateNextBlockPackage())
				return 0;
		}
	//	^- eklmn: bugfix(03): keeping user in queue if he doesn't need anything (v2)

	//	Feed data to this slot, one block at a time, as long as there's
	//	enough bandwidth left for the entire next block.
	//	There must also be blocks to send (not empty request queue)
		while (!m_blockSendQueue.IsEmpty() && m_blockSendQueue.GetHead()->GetRealPacketSize() + amounttransferred <= nMaxAmmount)
		{
		//	Get next block and give it to the m_pRequestSocket for transfer
			Packet * pPacket = m_blockSendQueue.RemoveHead();
			uint32 nBlockSize = pPacket->GetRealPacketSize();
		//	-khaos--+++>
		//	Extended statistics information based on which client software and which port we sent this data to...
		//	This also updates the grand total for sent bytes, etc.  And where this data came from.  Yeesh.
		//	eklmnf: we have to call function before packet will sended, otherwise packet will be deleted & we have Access violation
			g_eMuleApp.m_pGlobPrefs->Add2SessionTransferData(GetClientSoft(), GetUserPort(), pPacket->IsFromPF(), TRANSFERTYPE_UPLOAD, nBlockSize);
		//g_eMuleApp.UpdateSentBytes(nBlockSize);
			m_pRequestSocket->SendPacket(pPacket, true, false);
			pPacket = NULL;	//eklmn: bugfix(17): upload fix(reset pPacket in SendBlockData)
		//m_pRequestSocket->ResetTimeOutTimer();

		//	We've transferred some data. Store this in relevant statistics variables
			m_dwTransferredUp += nBlockSize;
			amounttransferred += nBlockSize;
		//	g_eMuleApp.m_pUploadQueue->AddUpDataRateMSOverhead(24, 160);
			m_pCredits->AddUploaded(nBlockSize, this->GetIP());

		//	If we're out of prepared blocks, create some more
			if (m_blockSendQueue.IsEmpty())
				CreateNextBlockPackage();
		}

	//	If this slot wasn't given enough data to empty it's request queue,
	//	it wants more data.

		if (!m_blockSendQueue.IsEmpty())
		{
			wantsMoreBandwidth = true;
		}

	//	Remember when this connection was last fed some data, so we
	//	can make sure we give it data often enough to not time out.
		if (amounttransferred > 0)
		{
			SetLastGotULData();
		}
	}
#endif //OLD_SOCKETS_ENABLED

//	Store how much data we've transferred this round,
//	to be able to calculate average speed later
	m_averageUDRList.push_front(m_dwTransferredUp);
	m_averageULTickList.push_front(curTick);

//	Don't store more than 40 seconds of statistics
//while(curTick-m_averageULTickList.back() > 40000) {
	while (m_averageULTickList.size() > 500)
	{
		m_averageUDRList.pop_back();
		m_averageULTickList.pop_back();
	}

//eklmn: since performance is more importatnt, that graphical output this function moved to the end
//	Check if it's time to update the display.
	if (curTick - m_dwLastRefreshedULDisplayTime > MINWAIT_BEFORE_ULDISPLAY_WINDOWUPDATE && m_averageUDRList.size() >= 2)
	{
	//	Calculate average data rate
	//	Present this slot's speed as 0 unless there's a good chunk
	//	of stored statistics for it. If there's to little data, the
	//	UL speed will be far off, so we might as well hide it in those cases.
		uint32 timePeriod = m_averageULTickList.front() - m_averageULTickList.back();
		if (timePeriod > 5000)
			m_lUpDataRate = static_cast<sint32>((static_cast<double>(m_averageUDRList.front() - m_averageUDRList.back())) * 1000.0 / timePeriod);
		else
			m_lUpDataRate = -1;

		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlUploadList.UpdateClient(this);
		m_dwLastRefreshedULDisplayTime = curTick + (uint32)(rand() / (RAND_MAX / 1000));	//SyruS moved rand out of if(...)
	}

	return amounttransferred;

	EMULE_CATCH
	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	FlushSendBlocks() sends the rest of the last block queued for upload to this client if it has been split across
//		multiple packets. Call this when you stop upload, or the request socket might be not able to send.
void CUpDownClient::FlushSendBlocks()
{
	EMULE_TRY

	bool		bBreak = false;

#ifdef OLD_SOCKETS_ENABLED
	while ( !m_blockSendQueue.IsEmpty() && m_blockSendQueue.GetHead()->IsSplit()
		 && m_pRequestSocket != NULL && m_pRequestSocket->IsConnected() && !bBreak )
	{
		Packet		*pPacket = m_blockSendQueue.RemoveHead();

		bool		bBreak = pPacket->IsLastSplit();

		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
		m_pRequestSocket->SendPacket(pPacket, true, false);
	}
#endif //OLD_SOCKETS_ENABLED
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendHashsetPacket(char* forfileid)
{
	EMULE_TRY
	CKnownFile * file = g_eMuleApp.m_pSharedFilesList->GetFileByID((uchar*)forfileid);
	if (!file)
	{
		AddDebugLogLine(false, GetResString(IDS_ERR_REQ_FNF) + CString(" (SendHashsetPacket)"));
		return;
	}

	CSafeMemFile data(1024);
	data.Write(file->GetFileHash(), 16);
	uint16 parts = file->GetHashCount();
	data.Write(&parts, 2);
	for (int i = 0; i < parts; i++)
		data.Write(file->GetPartHash(i), 16);
	Packet* packet = new Packet(&data);
	packet->m_eOpcode = OP_HASHSETANSWER;
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(packet, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}

double CUpDownClient::GetClientSpeed()
{
	EMULE_TRY
//	Calculates the session speed of the client
//	Providing the other Clients with this info
//	gives then the possibility to select the
//	sources more intelligently
	if (g_eMuleApp.stat_transferStarttime == 0)
		return 0.0;

	DWORD running = (GetTickCount() - g_eMuleApp.stat_transferStarttime) / 1000;
	if (running < 5)
		return 0.0;

	return static_cast<double>(g_eMuleApp.stat_sessionSentBytes / 1024.0) / running;
	EMULE_CATCH
	return 0.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ClearUploadBlockRequests()
{
	EMULE_TRY

	FlushSendBlocks();
	for (POSITION pos = m_blockRequestsQueue.GetHeadPosition();pos != 0;m_blockRequestsQueue.GetNext(pos))
		delete m_blockRequestsQueue.GetAt(pos);
	m_blockRequestsQueue.RemoveAll();

	for (POSITION pos = m_doneBlocksList.GetHeadPosition();pos != 0;m_doneBlocksList.GetNext(pos))
		delete m_doneBlocksList.GetAt(pos);
	m_doneBlocksList.RemoveAll();

	for (POSITION pos = m_blockSendQueue.GetHeadPosition();pos != 0;m_blockSendQueue.GetNext(pos))
		delete m_blockSendQueue.GetAt(pos);
	m_blockSendQueue.RemoveAll();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendRankingInfo()
{
	EMULE_TRY

#ifdef OLD_SOCKETS_ENABLED
	if (m_pRequestSocket == NULL)
		return;
#endif //OLD_SOCKETS_ENABLED

	if (!ExtProtocolAvailable())
		return;
	uint16 nRank = g_eMuleApp.m_pUploadQueue->GetWaitingPosition(this);
//	By Maverick
#ifndef DISABLE_EXTENDED_INFO_EXCHANGE
 #define DISABLE_EXTENDED_INFO_EXCHANGE
//	The extended info exchange is currently disabled since it is possible
//	to ask the client with several userhashes for the file and then use
//	the one with the best UL/DL ratio
//	It is planned to reactivate this feature, but with a protection against
//	this userhash spoofing.
//	At the last stage this faeture should make it possible to choose the
//	sources more intelligently by taking UL/DL ratio, queuerank and
//	speed of the other client into account.
#endif

#ifdef DISABLE_EXTENDED_INFO_EXCHANGE

	if (!nRank)
		return;
#else

	uint16 nBaseModifier = (uint16) GetScore(false, false, true); //16-Bits are enough for base value
	uint32 nScore = GetScore(false, false, false);

	uint8 nCommunity = -1;
	if (!g_eMuleApp.m_pGlobPrefs->CommunityEnabled())
		nCommunity = 2;
	else if (IsCommunity())
		nCommunity = 1;


	double dblBaseValue = 1.0;
	if (m_pCredits != (CClientCredits*)NULL)
	{
		dblBaseValue *= m_pCredits->GetScoreRatio(this->GetIP());
	}
	dblBaseValue *= 10;
	uint16 nRatio = (uint16)dblBaseValue;
#endif

	uint16 nClientSpeed = (uint16)((GetClientSpeed() < 65535) ? GetClientSpeed() : 65535);
//	Providing the other clients with this info
//	can not be used to leech

	Packet* packet = new Packet(OP_QUEUERANKING, 12, OP_EMULEPROT);
	memset2(packet->m_pcBuffer, 0, 12);
	memcpy2(packet->m_pcBuffer + 0, &nRank, 2);
#ifndef DISABLE_EXTENDED_INFO_EXCHANGE

	memcpy2(packet->m_pcBuffer + 2, &nBaseModifier, 1);		// UINT8 is enough 0-100
	memcpy2(packet->m_pcBuffer + 3, &nRatio, 2);			// UINT16 needed saw about 32000 max (SF * UL/DL * RF)
	memcpy2(packet->m_pcBuffer + 5, &nScore, 4);			// UINT32 needed
	memcpy2(packet->m_pcBuffer + 9, &nCommunity, 1);		// 2-Bits needed
#endif

	memcpy2(packet->m_pcBuffer + 10, &nClientSpeed, 2);		// UINT16 (I think not many have more than 65 MByte UL)
//	END Maverick
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(packet, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}

void CUpDownClient::SendCommentInfo(CKnownFile *file)
{
	EMULE_TRY
	if (!m_bCommentDirty || file == NULL || !ExtProtocolAvailable() || m_byteAcceptCommentVer < 1)
		return;
	m_bCommentDirty = false;

	EnumPartFileRating rating = file->GetFileRating();
	CString desc = file->GetFileComment();
	if (file->GetFileRating() == PF_RATING_NONE && desc.IsEmpty())
		return;

	CSafeMemFile data(256);
	data.Write(&rating, sizeof(rating));
	int length = desc.GetLength();
	if (length > 128)
		length = 128;
	data.Write(&length, sizeof(length));
	if (length > 0)
		data.Write(desc.GetBuffer(), length);
	Packet *packet = new Packet(&data, OP_EMULEPROT);
	packet->m_eOpcode = OP_FILEDESC;
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(packet, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}

void CUpDownClient::AddRequestCount(uchar* fileid)
{
	EMULE_TRY
	for (POSITION pos = m_requestedFilesList.GetHeadPosition();pos != 0;m_requestedFilesList.GetNext(pos))
	{
		Requested_File_Struct * cur_struct = m_requestedFilesList.GetAt(pos);
		if (!md4cmp(cur_struct->m_fileHash, fileid))
		{
			if (::GetTickCount() - cur_struct->m_dwLastAskedTime < MIN_REQUESTTIME)
			{
				if (GetDownloadState() != DS_DOWNLOADING)
					cur_struct->m_byteNumBadRequests++;
				if (cur_struct->m_byteNumBadRequests == BADCLIENTBAN) //katsyonak 16.06.03
					Ban(_T("AUTOMATED MESSAGE: You're being banned because your client is behaving aggressively!"), GetResString(IDS_BAN_AGGRESSIVECLI));
			}
			else
			{
				if (cur_struct->m_byteNumBadRequests)
					cur_struct->m_byteNumBadRequests--;
			}
			cur_struct->m_dwLastAskedTime = ::GetTickCount();
			return;
		}
	}
	Requested_File_Struct* new_struct = new Requested_File_Struct;
	memset2(new_struct, 0, sizeof(Requested_File_Struct));
	md4cpy(new_struct->m_fileHash, fileid);
	new_struct->m_dwLastAskedTime = ::GetTickCount();
	m_requestedFilesList.AddHead(new_struct);
	EMULE_CATCH
}

void CUpDownClient::UnBan()
{
	EMULE_TRY
	m_bBanned = false;
	m_dwBanTime = 0;
	SetWaitStartTime();
	g_eMuleApp.m_pUploadQueue->UpdateBanCount();
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();
//--- xrmb:hashthieves1 ---
	g_mapOffenseCounter.RemoveKey(GetUID());
	g_mapHashBase.RemoveKey(GetUID());
	m_bIsHashThief = false;
//--- :xrmb ---
	for (POSITION pos = m_requestedFilesList.GetHeadPosition();pos != 0;m_requestedFilesList.GetNext(pos))
	{
		Requested_File_Struct * cur_struct = m_requestedFilesList.GetAt(pos);
		cur_struct->m_byteNumBadRequests = 0;
		cur_struct->m_dwLastAskedTime = 0;
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);
	EMULE_CATCH
}

void CUpDownClient::Ban(LPCTSTR line, const CString& reason, ...)
{ //katsyonak 16.06.03
	EMULE_TRY

	if (!g_eMuleApp.m_pGlobPrefs->BanEnabled())
	{ // DbT: if the ban function is disable, no need to ban a client (default is enable!)
		m_bBanned = false;
		m_dwBanTime = 0;
		return;
	}
	if (m_eUploadState == US_UPLOADING)
	{	//SyruS kick already downloading client back to waitingqueue
		AddDebugLogLine(false, _T("Client '%s' banned and kicked out of uploadqueue %s"), GetUserName(), reason);
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BAN, true);
		g_eMuleApp.m_pUploadQueue->AddClientToQueue(this, true);
	}
	if (m_bBanned)
	{	//SyruS no need to ban a client (and send a message) again except for extending the bantime
		AddDebugLogLine(false, _T("Client '%s' is already banned from the uploadqueue %s"), GetUserName(), reason);
		m_dwBanTime = ::GetTickCount();
		return;
	}
	if (g_eMuleApp.m_pGlobPrefs->CommunityNoBanEnabled() && g_eMuleApp.m_pGlobPrefs->CommunityEnabled() && IsCommunity())
		return;
	m_bBanned = true;
	g_eMuleApp.m_pUploadQueue->UpdateBanCount();
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();
	m_dwBanTime = ::GetTickCount();
	if (g_eMuleApp.m_pGlobPrefs->IsBanMessageEnabled())
	{
		g_eMuleApp.m_pdlgEmule->m_wndChat.chatselector.StartSession(this, false);
		g_eMuleApp.m_pdlgEmule->m_wndChat.chatselector.SendMessage(line);
		g_eMuleApp.m_pdlgEmule->m_wndChat.chatselector.EndSession(this);		// now it automatically closes the session afterwards
		if (!g_eMuleApp.m_pGlobPrefs->GetCMNotLog())
			AddLogLine(false, IDS_SENTBAN, GetUserName(), reason);
	}
	else
		if (!g_eMuleApp.m_pGlobPrefs->GetCMNotLog())
			AddLogLine(false, IDS_CLIENTBLOCKED, GetUserName(), reason);

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}

void CUpDownClient::UDPFileReasked()
{
	EMULE_TRY
	AddAskedCount();
	SetLastUpRequest();
	SetLastL2HACExecution(); //<<-- enkeyDEV(th1) -L2HAC
	uint16 nRank = g_eMuleApp.m_pUploadQueue->GetWaitingPosition(this);
	Packet* response = new Packet(OP_REASKACK, 2, OP_EMULEPROT);
	memcpy2(response->m_pcBuffer, &nRank, 2);

//	By Maverick
	ProxySettings proxy = g_eMuleApp.m_pGlobPrefs->GetProxy();
	if (!proxy.m_bUseProxy)
	{
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(response->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

		g_eMuleApp.m_pClientUDPSocket->SendPacket(response, GetIP(), GetUDPPort());
#endif //OLD_SOCKETS_ENABLED

	}
	EMULE_CATCH
}

//	Lord KiRon
CString CUpDownClient::GetDownloadStateAsString(void)
{
	return GetStatusDLQueueString((EnumDLQState)GetDownloadState());
}
//	Lord KiRon - end
//	By Maverick
uint32	CUpDownClient::GetRemoteScore()
{
	EMULE_TRY
	if (IsRemoteInfoAvaiable())
		return m_dwRemoteScore;
	EMULE_CATCH
	return 0;	// We do not know for how long we are on the queue
//	Furhtermore the we do not know the remote ratio.
}

uint16	CUpDownClient::GetRemoteRatio()
{
	EMULE_TRY
	if (IsRemoteInfoAvaiable())
		return m_uRemoteRatio;
//	This calculation is not very accurate for several reasons:
//	- We do not know the file priority
//	- We do not save the client info if he has nothing uploaded to us
//	- Other factores like community etc

	/* // Myself Bot user?
	bool bIsBotuser;	//we still assume that the other client 'punishes' us if our name is 'wrong' 	//Cax2 - not anymore
	CString pszUsername = g_eMuleApp.m_pGlobPrefs->GetUserNick();
	if (pszUsername)
		bIsBotuser = (strstr(pszUsername, "edk-files") || strstr(pszUsername, "finder")) && strstr(pszUsername, "com");
	else bIsBotuser = false;
	*/

	double dblBaseValue = 100.0;
//if (bIsBotuser) dblBaseValue *= 0.9f;	//Cax2 - not anymore
#ifdef OLD_SOCKETS_ENABLED

	if (g_eMuleApp.m_pServerConnect->IsLowID())
		dblBaseValue *= 0.8;
#endif //OLD_SOCKETS_ENABLED

	if (GetDownloadState() == DS_DOWNLOADING)
		dblBaseValue *= 1.5;
	if (IsFriend())
		dblBaseValue *= 2.0;
	if (g_eMuleApp.m_pGlobPrefs->CommunityEnabled() && IsCommunity())
		dblBaseValue *= 2.0;
	dblBaseValue *= GetRemoteBaseModifier();
	dblBaseValue *= GetSmallFilePushRatio() ;
	dblBaseValue *= GetRareFilePushRatio() ;
	return static_cast<uint32>(dblBaseValue);

	EMULE_CATCH

	return 100;
}

double CUpDownClient::GetRemoteBaseModifier()
{
	EMULE_TRY
	if (IsRemoteInfoAvaiable())
		return static_cast<double>(m_uRemoteBaseModifier) / 100.0;
//	This calculation is not very accurate for some reasons:
//	- Crashes can cause that either client can not write
//	its client data. This can cause differences of the
//	client data at the two clients.
	if (m_pCredits == reinterpret_cast<CClientCredits*>(-1))
		m_pCredits = NULL;
	if (m_pCredits != NULL)
		return m_pCredits->GetScoreRatio(this->GetIP(), true);
	EMULE_CATCH
	return 1.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::GetCurrentlyUploadingPart()
{
	EMULE_TRY

	if (!m_doneBlocksList.IsEmpty())
	{
	//	Do we need to check upload state here?
		Requested_Block_Struct		*pFirstDoneBlock = m_doneBlocksList.GetHead();

	//	Calculate which part the block is in.
		uint32		dwPartNum = pFirstDoneBlock->m_dwStartOffset / PARTSIZE;

		return dwPartNum;
	}
	else
		return 0xFFFF;

	EMULE_CATCH

	return 0xFFFF;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetUploadState(EnumULQState news)
{
	m_eUploadState = news;
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.UpdateClient(this);
}
